! ESA (C) 2000-2018 
!
! This file is part of ESA's XMM-Newton Scientific Analysis System (SAS).
!
!    SAS is free software: you can redistribute it and/or modify
!    it under the terms of the GNU General Public License as published by
!    the Free Software Foundation, either version 3 of the License, or
!    (at your option) any later version.
!
!    SAS is distributed in the hope that it will be useful,
!    but WITHOUT ANY WARRANTY; without even the implied warranty of
!    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
!    GNU General Public License for more details.
!
!    You should have received a copy of the GNU General Public License
!    along with SAS.  If not, see <http://www.gnu.org/licenses/>.
module interpolate_mod
use dal
use sla
use cubicshepard_mod
use gridinterpolation_mod
use omallocation_mod
use omminim_mod
use omsmoothimage_mod
use brent_mod
use fitellipse_mod
public


type(structShepard), private, save :: shepard(10)
type(GridInterpolant), private, save :: gridData(10)
real(double), private :: minX = 0.0, maxX = 0.0
real(double), private :: minY = 0.0, maxY = 0.0


!*********************************************************
! For ellipse fitting
!
!********************************************************
real(double), private :: xCentre1=0.0, yCentre1=0.0
real(double), private :: cs = 0.0, sn = 0.0, height = 0.0


contains
!*********************************************************
!
!
!********************************************************
subroutine initialiseshepard(n, x, y, f, shepNo, minSep)
    integer, intent(in) :: n
    real(double), dimension(1 : n) :: x, y, f
    integer, intent(in) :: shepNo
    real(double), intent(in) :: minSep
!    integer :: i

    write(*, *) 

!    do i = 1, n
!        write(*, *) i, ") x=", x(i), " y=", y(i), " f=", f(i)
!    end do


    call shepardConstruct(shepard(shepNo), n, x, y, f, minSep)

 

end subroutine initialiseshepard
!***************************************************************
!
!
!**************************************************************/
real(double) function shepinterpolate(x, y, shepNo)
    real(double), intent(in) :: x, y
    integer, intent(in) :: shepNo
    shepinterpolate = ShepardInterpolate(shepard(shepNo), x, y)

end function shepinterpolate

!*****************************************************************
!
!
!***************************************************************** 
subroutine maximizer(nx, ny, f1, dw, maxDx, maxDy, maxValue)
    integer, intent(in) :: nx, ny
    real(double) :: f1(1 : nx * ny)
    real(double), intent(inout) ::  maxDx, maxDy 
    real(double), intent(inout) :: maxValue
    real(single), pointer, dimension(:) :: x=>null(), y=>null()
    real(single), pointer :: f(:, :) => null()
    real(double), intent(in) :: dw
    integer :: i, j, n
    real(single) :: xc, yc
    real(double) :: xFit(2), step(2), var(2)
    real(double) :: funcValue

!    real(double), pointer, dimension(:) :: xEllipse => null(), yEllipse => null()

!    real(double) :: x0,y0,xfoci1,yfoci1,xfoci2,yfoci2, &
!                    rmaj,rmin,foci_dist,tilt,eccen,    &
!                    a,asig,b,bsig,c,csig,d,dsig,e,esig,f2,fsig,chisq
      
    
    integer :: iFault
    
  

    x => allocateSingle(1, nx, "x")
    y => allocateSingle(1, ny, "y")
    f => allocate2dSingle(0, nx - 1, 0, ny - 1, "f")


    xc = real(nx) / 2.0 + 0.5
    yc = real(ny) / 2.0 + 0.5

!write(*, *) "maximizer- nx=", nx, " ny=", ny, " xc=", xc, " y=", yc

    minX = 10000.0
    maxX = -minX
  

    do i = 1, nx
        x(i) = (float(i) - xc) * dw
        minX = min(minX, dble(x(i)))
        maxX = max(maxX, dble(x(i)))

    end do

    minY = 10000.0
    maxY = -minY
 

    do i = 1, ny
        y(i) = (float(i) - yc) * dw
        minY = min(minY, dble(y(i)))
        maxY = max(maxY, dble(y(i)))

    end do

    n = 1

    do i = 0, nx - 1
        do j = 0, ny - 1
            f(i, j) = f1(n)
!            write(*, *) "x=", x(i+1), " y=", y(j+1), " f=", f1(n)
            n = n + 1
        end do
    end do
 
  
    call GridInterpolantConstruct(nx, ny, x, y, f, gridData(1))

    xFit(1) = maxDx
    xFit(2) = maxDy
    step = 0.005
    iFault = 0

    call omminim(xFit, step, 2, funcValue, 1000, 0.0000001_double, 4, 1, &
          0.00000001_double, var, maxFunction, iFault)

!    write(*, *) "FuncValue=", funcvalue, " ifault=", ifault
    if(ifault .eq. 0) then
        maxDx = xFit(1)
        maxDy = xFit(2)
        maxValue = -funcValue
!        write(*, *) "maxdx=", maxdx, " maxdy=", maxdy
    end if


    !*******************************************************************
    ! For test fit an ellipse round the maximum
    !******************************************************************


!    nEllipse = 360
!    xCentre1 = 0.0
!    yCentre1 = 0.0


    !*****************************************************************
    ! Find the mean correlation coefficient at a radius of 0.5 pixels
    ! from the centre
    !
    !****************************************************************
!    rMin = 0.5
!    height = 0.0_double
       
!    do i = 1, nEllipse, -1
!        angle = dble(i) * degreesToRadians
!        cs = cos(angle)
!        sn = cos(angle) 

!        xc = xCentre1 + rMin * cs
!        yc = yCentre1 + rMin * sn
!        height = height + dble( GridInterpolate1(xc, yc,  gridData(1)) )
    
!    end do
!    height = height / dble(nEllipse)

!    write(*, *) "Computed mean correlation at a radius of 0.5 pixels=", height

!    xEllipse => allocateDouble(1, nEllipse, "xEllipse")
!    yEllipse => allocateDouble(1, nEllipse, "yEllipse")



!    write(*, *) "Fitting ellipse to height=", height 

!    minRadius = 0.05
!    maxRadius = 0.95
!    tol = 0.0000000001_double
!    n = 0
!    do i = 1, nEllipse, -1
!        angle = dble(i) * degreesToRadians
!        cs = cos(angle)
!        sn = cos(angle) 

!        do j = 0, 9
!            minRadius = real(j) * 0.1  
!            maxRadius = minRadius + 0.05

            !*************************************************************
            ! Determine the radius at which the correlation equals height
            !
            !***********************************************************
!            okay = zbrac(minRadius, maxRadius, 20, ellipseFitFunc)
!            if(okay .and. minRadius .gt. 0.0 .and. minRadius .lt. 0.8) then
!                write(*, *) "Line 207 okay=true,  minr=", minRadius, " max=", maxRadius  
!                exit
!            end if
           
!        end do
!        if(okay) then  
!            maxRadius = min(1.0_double, minRadius) 
!            fMin = BRENT(minRadius, maxRadius, tol, rMin, ellipseFitfunc)
!            if(i < 10) then
!                write(*, *) "Brent fit- fmin=", fmin, " rmin=", rmin
!            end if 
!            n = n + 1
!            xEllipse(n) = xCentre1 + rMin * cs
!            yEllipse(n) = yCentre1 + rMin * sn
!             write(*, *) i, ") angle=", angle * radiansToDegrees, " x=", xEllipse(n), " y=", yEllipse(n) 

!        end if

!    end do
!    if (n .ge. 100) then
         
!        call fitellipse(xEllipse, yEllipse, n,                &
!                       x0,y0,xfoci1,yfoci1,xfoci2,yfoci2, &
!                       rmaj,rmin,foci_dist,tilt,eccen,    &
!                       a,asig,b,bsig,c,csig,d,dsig,e,esig,f2,fsig,chisq)
      
!        write(*, *) "Fitted ellipse parameters- chisq=", chisq
!        write(*, *) "xcentre=", x0, " ycentre=", y0
!        write(*, *) "semia=", rmaj, " semib=", rmin, " tilt=", tilt
 
!       if(rMaj .gt. 0.0) then

!       end if

!    end if
!    deallocate(xEllipse, yEllipse)
!    nullify(xEllipse, yEllipse)


    call GridInterpolantDestruct(gridData(1))
    


    deallocate(x, y, f)
    nullify(x, y, f)

end subroutine maximizer


!*********************************************************
! The function to be minimised
!*********************************************************
real(double) Function maxFunction(x)
    real (double), intent(inout)  :: x(:)

    if(x(1) .lt. minX .or. x(1) .gt. maxX .or. &
       x(2) .lt. minY .or. x(2) .gt. maxY) then

        maxFunction = 1e6
        return
    end if

    maxFunction = -dble( GridInterpolate1(real(x(1)), real(x(2)),  gridData(1)) )

!    write(*, *) "maxfunction- x=", x(1), " y=", x(2), " f=", maxFunction


end function maxFunction
!*******************************************************
!
!
!
!******************************************************
real(double) function ellipseFitfunc(radius)
    real(double), intent(in) :: radius
    
    real(double) :: x, y, f
    integer, save :: count = 0
    x = xCentre1 + radius * cs
    y = yCentre1 + radius * sn

    f = dble(GridInterpolate1(real(x), real(y),  gridData(1) ))


    ellipseFitFunc = height - f

    if(count .lt. 0) then
        count = count + 1
       write(*, *) "Radius=", radius, " x=", x, " y=", y, " funcvalue=", ellipseFitFunc, &
          " height=", height, " f=", f
    end if 

end function ellipseFitfunc


!*****************************************************************
!
!
!***************************************************************** 
!subroutine smoothimage(nx, ny, f1, maxDx, maxDy, maxValue)
!    integer, intent(in) :: nx, ny
!    real(double) :: f1(1 : nx * ny)
!    real(single)


end module interpolate_mod
