// ESA (C) 2000-2018 
//
// This file is part of ESA's XMM-Newton Scientific Analysis System (SAS).
//
//    SAS is free software: you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation, either version 3 of the License, or
//    (at your option) any later version.
//
//    SAS is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with SAS.  If not, see <http://www.gnu.org/licenses/>.
#include "ommosaic.h"
#include<iostream>
#include "sourcelist.h"
#include "utility.h"
#include <f90cxx.h>
#include<numeric>
#include"astrometrydefs.h"
#include "statistics.h"
#include "Dal.h"
#include "errstr.h"
#include<vector>
#include<map>
#include<cstdlib>
#include "interpolation.h"
#include "omwcs.h"
#include "omstackprototypes.h"
//gcc4.8.2 portability changes
#include <cstdio>


/**************************************************************
*
*
* The main program
*
*
**************************************************************/
int imageType = 1;  // 0=sky, 1=normal (obtained from first input image)

bool mixedBinning = false; // Records whether or not images have different binning factors

void ommosaic()
{

    
    /***************************************************************
    *
    *
    *
    ****************************************************************/          
    int noStackedFiles  = 0;   
    vector<string> stackedFileList;
    vector<OMWCS> stackedWcsList; 
    
    float minCorr = realParameter("mincorr"); 


    /******************************************************************
    *
    * Store image files in a map
    *
    ******************************************************************/
    map<string, vector<string> >     fileListMap; 
    map<string, vector<OMWCS> >      wcsListMap; 


    //printf("L56 minCorr=%f \n",minCorr);

    OMWCS wc3; // The parameters for the mosaiced image

    if( abs(minCorr) < 0.00001)   // Don't want to align the images- just stack them
    {
  

        loadData(fileListMap, wcsListMap); 

	
	//cout << "L65 Size of fileListMap: " << fileListMap.size() << '\n';
 	//cout << "L66 Size of wcsListMap: " << wcsListMap.size() << '\n';
	

        map<string, vector<string> >::iterator iter;
        for(iter = fileListMap.begin(); iter != fileListMap.end(); ++iter)
	{  
	    string obsId = iter->first;
	  
	    vector<string> fileList = iter->second; 
	        
	    //cout << "L75 obsId: " << obsId << '\n';
	    //cout << "L76 fileList: " << fileList << '\n';	    

	    //cout << "L81 -----------"  << '\n';
	    // *******************************8
	    // 2011-12-12 here the RA data should be checked for consistency
	    // of all xLeft and xRight

	    /*
            for(unsigned i = 0; i < fileList.size(); i++)
	    {
	      cout << "L87 "<< i << ") " << " xleft=" << setprecision(6) <<  
		setiosflags(ios::fixed)   << wcsListMap[obsId][i].data->xLeft * radiansToDegrees << 
		" xright=" <<  setprecision(6) <<  setiosflags(ios::fixed)   
		   << wcsListMap[obsId][i].data->xRight * radiansToDegrees << endl; 	      
	    }	   
	    */ 

	    for(unsigned i = 0; i < fileList.size(); i++)
	    { 
	      if (wcsListMap[obsId][i].data->xLeft * radiansToDegrees <0.0 ||
		  wcsListMap[obsId][i].data->xRight * radiansToDegrees <0.0){
		wcsListMap[obsId][i].data->xLeft+=360.0/radiansToDegrees;
		wcsListMap[obsId][i].data->xRight+=360.0/radiansToDegrees;
	      }
	    }

	    /*
            for(unsigned i = 0; i < fileList.size(); i++)
	    {
	      cout << "L104 "<< i << ") " << " xleft=" << setprecision(6) <<  
		setiosflags(ios::fixed)   << wcsListMap[obsId][i].data->xLeft * radiansToDegrees << 
		" xright=" <<  setprecision(6) <<  setiosflags(ios::fixed)   
		   << wcsListMap[obsId][i].data->xRight * radiansToDegrees << endl; 	      
	    }	   
	    */

	    //**************************
            wc3.setImageType( imageType );
          

            if(imageType == 0)
	    {
                checkAstrometry( wcsListMap[obsId], wc3 );
                updateImageHeaders( fileListMap[obsId], wcsListMap[obsId]);
	    }

	    //cout << "L99 fileList.size: " << fileList.size() << '\n';
	    //cout << "L100 -----------"  << '\n';
            for(unsigned i = 0; i < fileList.size(); i++)
	    {

	      //cout << " i=" << i << ") crval1=" << wcsListMap[obsId][i].data->crval1 * radiansToDegrees 
	      //   << " frameTime=" << wcsListMap[obsId][i].data->frameTime << endl;

	      /*
	      cout << i << ") L106 crval1=" << setprecision(6) <<  setiosflags(ios::fixed) 
		   << wcsListMap[obsId][i].data->crval1 * radiansToDegrees << 
		" crval2=" << setprecision(6) <<  setiosflags(ios::fixed)  << 
		wcsListMap[obsId][i].data->crval2 * radiansToDegrees << 
		" xleft=" << setprecision(6) <<  setiosflags(ios::fixed)   
		   << wcsListMap[obsId][i].data->xLeft * radiansToDegrees << 
		" xright=" <<  setprecision(6) <<  setiosflags(ios::fixed)   << 
		wcsListMap[obsId][i].data->xRight * radiansToDegrees << 
		" ytop=" <<  setprecision(6) <<  setiosflags(ios::fixed)   << 
		wcsListMap[obsId][i].data->yTop * radiansToDegrees << 
		" ybottom=" <<  setprecision(6) <<  setiosflags(ios::fixed)  << 
		wcsListMap[obsId][i].data->yBottom * radiansToDegrees << endl; 
	      */
   
                stackedFileList.push_back(fileListMap[obsId][i] );  
                stackedWcsList.push_back(wcsListMap[obsId][i] );



		// cout << i << ") Line 92 crval1=" << setprecision(6) <<  setiosflags(ios::fixed) << stackedWcsList[i].data->crval1 * radiansToDegrees << 
	        //           " crval2=" << setprecision(6) <<  setiosflags(ios::fixed)  << stackedWcsList[i].data->crval2 * radiansToDegrees << 
                //           " xleft=" << setprecision(6) <<  setiosflags(ios::fixed)   << stackedWcsList[i].data->xLeft * radiansToDegrees << 
                //           " xright=" <<  setprecision(6) <<  setiosflags(ios::fixed)   << stackedWcsList[i].data->xRight * radiansToDegrees<< 
                //           " ytop=" <<  setprecision(6) <<  setiosflags(ios::fixed)   << stackedWcsList[i].data->yTop * radiansToDegrees << 
	        //           " ybottom=" <<  setprecision(6) <<  setiosflags(ios::fixed)  << stackedWcsList[i].data->yBottom * radiansToDegrees  << endl; 

            }
        }

        noStackedFiles = (int)stackedFileList.size();
	//cout <<"L122 noStackedFiles= " << noStackedFiles << '\n';
    }
    else
    {
         
      
      //cout << "L130 trying to align images \n  ";

        /************************************************************************
	*
        * Try to accurately align the images
        *
        *************************************************************************/    

        map<string, vector<Alignment> >  alignmentListMap;

      
        initialise(fileListMap, wcsListMap, alignmentListMap);

    
        /***********************************************************
	* Loop through each obs-id and align the images from that obs-id
        *
        ************************************************************/

        map<string, vector<string> >::iterator iter;


        vector<OMWCS> wcsList;
        vector<string> fileList;
    
        for(iter = fileListMap.begin(); iter != fileListMap.end(); ++iter)
	{ 

	    string obsId = iter->first;
            if( fileListMap[obsId].size() > 1)
	    {           
                alignImages(alignmentListMap[obsId], 
                            fileListMap[obsId], 
                            wcsListMap[obsId],
                            stackedFileList, 
                            stackedWcsList,
                            noStackedFiles);       
            }
            else
	    {

           
	      fileList.push_back( fileListMap[obsId][0] );	 
              wcsList.push_back( wcsListMap[obsId][0] );

          

            }
           

        }       

   
        if(fileList.size() > 0)
	{

       
            vector<Alignment>  alignmentList;


            setAlignmentParameters(fileList, 
                                   wcsList, 
                                   alignmentList);
          

            alignImages(alignmentList, 
                        fileList, 
                        wcsList,
                        stackedFileList, 
                        stackedWcsList,
                        noStackedFiles);       
    

        }

        wc3.setImageType( imageType );
     
 
   }

    /**************************************************************
    *
    * Stack the files
    *
    ***************************************************************/
    if(stackedFileList.size() > 0)
    {

        /***********************************************************
        *
        * Stack the files
        *
        ************************************************************/
        vector<vector<float> > stackedImage, stackedExposureImage;
        vector<vector<int16> > stackedQualityImage;

    
  
	//			        OMWCS wc3( stackedWcsList[0] );
   
        string outputFile = stringParameter( "mosaicedset" );

	// 2014-02-02 vny: calculating frameTameMax and deadFracMax 
	double deadFracMax=0.0;
	double frameTimeMax=0.0;

        //stackFiles(stackedFileList, stackedWcsList, stackedImage, 
        //       stackedExposureImage, stackedQualityImage, wc3);
        stackFiles(stackedFileList, stackedWcsList, stackedImage, 
		   stackedExposureImage, stackedQualityImage, wc3,
		   deadFracMax, frameTimeMax);
	//cout << "L280 deadFracMax=" << deadFracMax << " frameTimeMax=" << frameTimeMax << endl; 
        /***************************************************************
        *
        * Create the output stacked-image file
        *
        ***************************************************************/
	
        getMeanPhotometryConstants(stackedWcsList, wc3);

        wc3.data->setDates(stackedFileList);
 
        wc3.data->createOutputFile(outputFile, stackedImage, 
                               stackedExposureImage, stackedQualityImage, 
				   //stackedFileList);
				   stackedFileList, deadFracMax, frameTimeMax);
    }
    

    /**************************************************************
    *
    * Only for images that have been undistorted
    *
    ***************************************************************/

    //   if(imageType == 1) 
    // {
    //    removeTemporaryFiles();
    //  }

}

/******************************************************************
*
*
* Subroutine for finding the offsets, dx and dy between
* the 2 input images
*
* NB Trying to align image 2 with image 1
*
*******************************************************************/
void process(string file1, OMWCS &wcs1, ImageData &data1, 
             string file2, OMWCS &wcs2, ImageData &data2, bool &okay, 
             double &maxR, int noStacked, double &maxDx, double &maxDy)
{

  
    errstr << "Cross-correlating " << file2 << " with " << file1
           << Sas::message(Msg::App, Msg::Sparse);


  
    /***************************************************************
    *
    * The following are used in the cross-correlation to 
    * speed things up
    *
    *
    ****************************************************************/
    setUsablePixels( data1 );

    float minCorr = realParameter( "mincorr" );
    int minNumPixels = intParameter("minnumpixels");
    if(data1.numberGoodPixels < minNumPixels)
    {
        errstr << "Unable to align images- too few pixels in image " << data1.fileName
                           << endl << "Number of pixels above the background=" << data1.numberGoodPixels << Sas::message(Msg::App, Msg::Sparse);


  
        okay = false;
       
    }

    setUsablePixels( data2 );


    if(data2.numberGoodPixels < minNumPixels)
    {
        errstr << "Unable to align images- too few pixels in image " << data2.fileName
                           << endl << "Number of pixels above the background=" << data2.numberGoodPixels << Sas::message(Msg::App, Msg::Sparse);

  
        
        okay = false;
  
        return;
    }


    removeBackground( data2 );

    /***************************************************************
    *
    * Compute the correlation coefficient between the two images
    * for a range of shifts dx and dy
    *
    ***************************************************************/

   

    maxDx = 0.0, maxDy = 0.0;

    maxR = 0.0; // No correlation
    double minR = 0.0;
    int numIntervals = intParameter("numintervals");

    int di = intParameter("maxdx") * numIntervals;



    double dw = 1.0 / double(numIntervals);
    int di1 = 2 * di + 1;
    vector<vector<double> > rImage( di1, vector<double>(di1, 0.0));

    
  
    crossCorrelate(  wcs1, data1, wcs2, data2.image, di, dw, rImage,  
                     maxDx, maxDy, minR, maxR);

  
    /************************************************************************
    *
    * Now have an initial estimate of the x and y offsets (maxDx and maxDy)
    *
    *
    * Smooth the correlation map image a little bit
    *
    ********************************************************/
    smoothImage( rImage );


    /*****************************************************************************
    *
    * Obtain a better estimate of the position of the peak correlation
    *
    ******************************************************************************/
    int numPeaks = 0;
    findCorrelationPeak(rImage, di, dw, maxDx, maxDy, maxR, numPeaks);
    

    if(minCorr > 0.0 && maxR < minCorr)
    {
     
        maxR = 0.0;
        okay = false;
        return; 
    }



    /***************************************************************
    *
    * Refine the correlation map centred around the peak
    *
    ***************************************************************/
    di = intParameter("di");  // was 50

    int size = 2 * di + 1;
    rImage.resize(size);

  
    for(unsigned i = 0; i < rImage.size(); i++)
    {
        rImage[i].resize(size);
        for(unsigned j = 0; j < rImage[i].size(); j++)
            rImage[i][j] = -1.0;
    }       
   

    maxR = 0.0;
    dw = 2.0 / double(di); // was 2.0
    minR = 100.0;

  
    /*************************************************************************
    *
    * Compute the cross-correlation on a finer grid
    * 
    *************************************************************************/
    crossCorrelate(  wcs1, data1, wcs2, data2.image, di, dw, rImage,  
                     maxDx, maxDy, minR, maxR);


    /*******************************************************
    *
    * Smooth the correlation map image a little bit
    *
    ********************************************************/
    smoothImage( rImage );


    /*****************************************************************************
    * Refine position of correlation peak
    *
    ******************************************************************************/
    double dx = 0.0, dy = 0.0;
    numPeaks = 0;
    findCorrelationPeak(rImage, di, dw, dx, dy, maxR, numPeaks);



    //     cout << "line 423 number of peaks=" << numPeaks << endl;


    if(numPeaks > 1 && minCorr > 0.0)
    {

       
        maxR = 0.0;
        okay = false;
        return; 
    }
    
    if(numPeaks == 0 && minCorr > 0.0)
    {
        maxR = 0.0;
        okay = false;
        return; 
    }
    

    /***************************************************
    *
    *
    *
    ****************************************************/
    refineCorrelation(rImage, dw, dx, dy, maxR);
    maxDx += dx;
    maxDy += dy;  

   
    if( minCorr > 0.0 && maxR < minCorr )
    {
        maxR = 0.0;     
        okay = false;
        return; 
    }

    okay = true;

    /***********************************************************************
    *
    * Need to add the maxDx and maxDy values to crpix1 and crpix2 of second ime
    *
    * maxDx is the x offset added to image 2 x coordinates that gave the best correlation.
    * maxDy is the y offset added to image 2 y coordinates that gave the best correlation.
    * These values need to be converted to delta ra and dec, respectively
    *
    * NB a positive maxDx means a negative delta ra
    * maxDx and maxDy are in pixels
    * 
    ************************************************************************/
    wcs2.data->updateReferencePixelCorrection( wcs1, maxDx, maxDy );
 
 
    /***************************************************************
    *
    *
    ***************************************************************/
    if(stringParameter("correlset").length() < 3) return;



    if(noStacked == 0)
    {
        createCorrelationFile(rImage, data1.fileName, data2.fileName, abs(maxR), maxDx, maxDy); 
    }
    else
    {
        updateCorrelationFile( rImage, data1.fileName, data2.fileName, abs(maxR), maxDx, maxDy );
    }

}

/*************************************************************************
*
* This function computes the correlation coefficient (0.0 to 1.0) for
* the two images.
* dx is the offset added to image 2 x coordinates
* dy is the offset added to image 2 y coordinates
*
*
*
***************************************************************************/
void crossCorrelate(  OMWCS &c1, ImageData &data1,
                      OMWCS &c2,                    
                      vector<vector<float> >  &image2,
                      int di, double dw,
                      vector<vector<double> >  &rImage,   // The correlation map image
                      double &maxDx, double &maxDy, double &minR, double &maxR)
{

    double dx = maxDx, dy = maxDy; 
    
    int nx2 = (int) image2.size();
    int ny2 = (int) image2[0].size();

    float xMax2 = (float)nx2 + 0.5;
    float yMax2 = (float)ny2 + 0.5;

   
    int minNumPixels = intParameter("minnumpixels");


    map<int, map<int, double> > sum1Map, sum2Map, sum3Map;
    map<int, map<int, int> >  numPixelsMap;


    /**********************************************************
    *
    * NB xOffset will go from -di to +di
    * and yOffset from -di to +di
    *  
    **********************************************************/
    vector<int> xOffset, yOffset;
    setUpOffsets( di, xOffset, yOffset);
 

    /*************************************************************
    *
    *
    * Loop through all the usable pixels
    *
    *************************************************************/
    minR = 100.0, maxR = -100.0;

    
    /********************************************************************
    *
    * NB the following is experimental to allow for different binnings
    *
    ********************************************************************/

    //    double factor = (double)c1.data->binAxis / (double)c2.data->binAxis;

    Interpolation<double> interpolator( image2 );

    for(unsigned i = 0; i < data1.x.size(); i++)
    {      
        
        /******************************************************************
	*
	* 
        *
        *******************************************************************/
        double x20 = 0.0, y20 = 0.0;
        c1.data->convertX1Y1ToX2Y2(data1.x[i], data1.y[i], c2, x20, y20);

        x20 += dx; // * factor;  // Add the x-shift
        y20 += dy; //  * factor;  // Add the y-shift
       

        /*********************************************************************
	*
        * Compute the correlation coefficient for a range of x and y offsets
        * centred on x2, y2
        *
        *********************************************************************/

        for(unsigned j = 0; j < xOffset.size(); j++)
	{
	    double x2 = (double)xOffset[j] * dw + x20;  // x20 is the x shift
	    double y2 = (double)yOffset[j] * dw + y20;  // y20 is the y shift

            if(x2 < 0.5) continue;
            if(x2 > xMax2) continue;
        
            if(y2 < 0.5) continue;
            if(y2 > yMax2) continue;
 
            double f2 = interpolator.BilinearInterpolation(x2, y2);
            /*********************************************************
	    *
            * Check for null or negative value
            *
            ***********************************************************/          
            if ( !isNull( f2 ) && f2 > 0.0 ) 
	    {     
                int x = xOffset[j];
                int y = yOffset[j];
                ++numPixelsMap[x][y]; // The number of pixels used
                sum1Map[x][y] += data1.f[i] * f2;	       
                sum2Map[x][y] += data1.fSquared[i];
                sum3Map[x][y] += f2 * f2; 
		 
                                  
            }

        }       // End of looping through the offsets

    }

    /**************************************************
    *
    * Find the offsets that gave the maximum correlation
    *
    ***************************************************/
    map<int, map<int, int> >::iterator iter;
   
    int count = 0;
    for(iter = numPixelsMap.begin(); iter != numPixelsMap.end(); ++iter)
    {
        int x = iter->first;
        map<int, int> numMap = iter->second;
        map<int, int>::iterator iter1;

        for(iter1 = numMap.begin(); iter1 != numMap.end(); ++iter1)
        {

            int y = iter1->first;
	    int numPixels = iter1->second;
            double r = 0.0;
 
	    if(numPixels >= minNumPixels)
	    {	       
	        if( sum2Map[x][y] > 0.0 && sum3Map[x][y] > 0.0)
		{
	            r = sum1Map[x][y] / sqrt( sum2Map[x][y] * sum3Map[x][y] );
                    if(r > maxR) // Record where the new maximim is
		    {
		        maxR = r;
                        maxDx = (double)x * dw + dx; // The new best x displacement
                        maxDy = (double)y * dw + dy; // The new best y displacement		        
		    }
                    else if( r < minR)
		    {
		        minR = r;
                    }
                 
                }
                    
            }
	    //	    cout << "x offset=" << x << " y=" << y << " r=" << r << " num pixels=" << numPixels << endl;
	    //  if( count%100 == 0)
	    // {
	    //  cout << " max R=" << maxR << " at dx=" << maxDx << " dy=" << maxDy << " number of pixels used=" << numPixels << endl;
	    // }
            ++count;
	    rImage[x + di][y + di] = r; // Set the correlatio image pixel
  
        }
    } // End of looping through the list of pixles    
}

/*********************************************************
*
*
*
*
*
*********************************************************/
double Mean(vector<vector<float> > &image)
{

    double sum = 0.0;
    int count = 0;
    for(unsigned i =0; i < image.size(); i++)
    {
        for(unsigned j =0; j < image[i].size(); j++)
	{
	  if (!isNull( image[i][j] ) )
	  {
	      ++count;
              sum += double(image[i][j] );
          }
        }
    }
    if(count > 0)
        return sum / double(count); 
    else
      return 0.0;
}

/*********************************************************
*
*
*
*********************************************************/
double Mean(vector<vector<float> > &image, int il, int iu, int jl, int ju)
{

    double sum = 0.0;
    int count = 0;
    for(int i = il; i <= iu; i++)
    {
    
        for(int j = jl; j <= ju; j++)
	{
	  if (!isNull( image[i][j] ) )
	  {
	      ++count;
              sum += double(image[i][j] );
          }
        }
    }
    if(count > 0)
        return sum / double(count); 
    else
        return 0.0;
}

/***********************************************************************
*
*
*
***********************************************************************/
bool isSkyImage( string file, OMWCS &c )
{
 
  //  cout << "line 743 isskyimage file=" << file << endl;
  //   c.data = new WcsData1;
  //  cout << "Line 725\n";

    return c.data->load( file );
   
}
    /*************************************************************************************
*
*
*
**************************************************************************************/
vector<vector<float> >  computeBackgroundImage(vector< vector<float> > &image, int dx)
{

    int nx = (int)image.size();
    int ny = (int)image[0].size();
    
    vector<float> pixel;
    for(int i = 0; i < nx; i++)
    {
        for(int j = 0; j < ny; j++)
        {
	  if ( !isNull(image[i][j] ) )
          {
              pixel.push_back(image[i][j] );
          }
        }
    }
 

    float globalMedian = Median(pixel, true, true);      

    
    pixel.clear();
    //    cout << "Global median=" << globalMedian << endl;
    // cout << "Computing background image nx=" << nx << " ny=" << ny << endl;;
    vector<vector<float> > backgroundImage(nx, vector<float>(ny));

    if((nx * ny) > 1)
    { 
        for(int i = 0; i < nx; i++)
        {
            for(int j = 0; j < ny; j++)
            {
	        backgroundImage[i][j] = globalMedian;
            }
        }
    }
    else
    {
        for(int i = 0; i < nx; i++)
        {   
	  // if( i % 100 == 0) cout << "Row " << i << endl;
            for(int j = 0; j < ny; j++)
	    {
	        float median = computeBackground(image, nx, ny, i, j, dx);
                if(median >= 0.0)
                    backgroundImage[i][j] = median;
                else
                    backgroundImage[i][j] = globalMedian;
           
            }
        }
    }
  
    return backgroundImage;

}
/************************************************************************
*
*
*
**********************************************************************/
float computeBackground(vector<vector<float> > &image, 
                    int nx, int ny, 
                    int i0, int j0, int dx)
{
    int il = max(0, i0 - dx);
    int iu = min(nx - 1, i0 + dx);

    int jl = max(0, j0 - dx);
    int ju = min(ny - 1, j0 + dx);

    vector<float> pixel;
    for(int i = il; i <= iu; i++)
    {
        for(int j = jl; j <= ju; j++)
        {
	    if ( !isNull(image[i][j] ) )
            {
                pixel.push_back(image[i][j] );
            }
        }

    }
    if(pixel.size() > 50)
    {
      return Median(pixel, true, true);      

    }
    else
    {
      return -1.0;

    }

}

/***************************************************************************
*
*
*
*
****************************************************************************/

void loadWcsData(string file, OMWCS &wcs)
{

  // cout << "Line 862 loadwcsdata\n"; 
    if (!isSkyImage( file, wcs ) )
    {
 
      //     Sas::fatal("Not a sky image-", file);


    }

  
}

/***************************************************************************
*
*
*
*
*
****************************************************************************/
void loadImageData(string file, ImageData &data)
{

    /**************************************************************
    *
    * Load the image
    *
    ***************************************************************/
    int nx = 0, ny = 0;
    data.fileName = file;    
    data.image = loadFloatImage( file, "PRIMARY", nx, ny);


    data.nx = nx;
    data.ny = ny;
 
    /*************************************************************
    *
    * Compute the background image
    *
    **************************************************************/

    int dr = 5;
    data.backgroundImage = computeBackgroundImage(data.image, dr);

} 
/*****************************************************************************
*
*
*
*
*****************************************************************************/
void stackFiles(vector<string> &fileList, vector<OMWCS> &wcsList,
                vector< vector<float> > &stackedImage, 
                vector< vector<float> > &stackedExposureImage,
                vector< vector<int16> > &stackedQualityImage,
                //OMWCS &c3)
		OMWCS &c3, double &deadFracMax, double &frameTimeMax)
  
{
  
  string outputFilename = stringParameter( "mosaicedset" );
  //cout << "===========================\n";
  errstr << "Stacking the images to "<<   outputFilename 
	 << Sas::message(Msg::App, Msg::Sparse);
  //cout << "L964 wcsList.size=" <<  wcsList.size() << endl;
  int maxBinAxis = 0;  // Find the maximum binning factor
  for(unsigned i = 0; i < wcsList.size(); i++){
    //cout << "L968 i=" << i << " --- " << endl;
    /************************************************************************
     * In the case of Sky images the corner coordinates will be RA and Dec
     * 
     * For unrotated images the corner coordinates will be full frame coordinates
     ************************************************************************/
    wcsList[i].data->correctReferencePixel();
    /* nbnbnbnbn 2013-08-16 */
    //wcsList[i].data->getCornerCoords();
    //printf("L976 xLefti=%f xRighti=%f \n", wcsList[i].data->xLeft,wcsList[i].data->xRight);
    /* nbnbnbnbn 2013-08-16 */

    /**********************************************************************
     *
     * Find the maximum value of binAxis
     *
     **********************************************************************/
    //cout << "L981 i=" << i << " wcsList.binAxis " << wcsList[i].data->binAxis << '\n';
    if ( wcsList[i].data->binAxis > maxBinAxis){
      maxBinAxis = max(maxBinAxis, wcsList[i].data->binAxis);
      
      c3.data->cdelt1 = wcsList[i].data->cdelt1;
      c3.data->cdelt2 = wcsList[i].data->cdelt2;
      c3.data->binAxis = maxBinAxis;
      
    }

  }// for(unsigned i = 0; i < wcsList.size()
  
  //cout << "L994 maxBinAxis= " << maxBinAxis << '\n';
  
  int binAxis = intParameter("binaxis");
  
  //cout << "L998 binAxis= " << binAxis << '\n';
  
  if( binAxis == 0 )
    {
      binAxis = maxBinAxis; // Use the maximum binning factor of all the files  
      
    }
  //cout << "L1005 using  binAxis=" << binAxis << endl;
  if (binAxis != c3.data->binAxis)
    {
      if(binAxis == 1)  // High resolution
	{
	  //cout << "L1012 high resolution \n";
	  c3.data->cdelt1 /= 2.0;
          c3.data->cdelt2 /= 2.0;
        }
      else              // Low resolution
	{
	  //cout << "L1018 low resolution \n";
	  c3.data->cdelt1 *= 2.0;
	  c3.data->cdelt2 *= 2.0;
	}
      c3.data->binAxis = binAxis;
    }
  
  //cout << "L1025 c3.binAxis=" << c3.data->binAxis << endl;
  if(c3.imageType == 0) //Sky image
    {
      
      //printf("L1026 xLeft=%f xRight=%f \n",wcsList[0].data->xLeft,wcsList[0].data->xRight);
      c3.data->xLeft = wcsList[0].data->xLeft;
      c3.data->xRight = wcsList[0].data->xRight;
      c3.data->yBottom = wcsList[0].data->yBottom;
      c3.data->yTop = wcsList[0].data->yTop;        
    }
  else
    {
      //printf("L1034 xLeft=%f xRight=%f \n",wcsList[0].data->xLeft,wcsList[0].data->xRight);
      c3.data->xLeft = wcsList[0].data->xLeft;
      c3.data->xRight = wcsList[0].data->xRight;
      c3.data->yBottom = wcsList[0].data->yBottom;
      c3.data->yTop = wcsList[0].data->yTop;
      c3.data->crpix1 = wcsList[0].data->crpix1;
      c3.data->crpix2 = wcsList[0].data->crpix2;        
    }
  
  /******************************************************************
   *
   * The maximum extent of the stacked image
   *
   *******************************************************************/
  
  //cout << "L1050\n";
  
  //cout << 0 << ") c3 xleft=" << setprecision(6) <<  setiosflags(ios::fixed) 
  //   << c3.data->xLeft  * radiansToDegrees<<  " xright=" << c3.data->xRight * radiansToDegrees 
  //   << " yBottom=" << c3.data->yBottom * radiansToDegrees << " yTop=" 
  //   << c3.data->yTop * radiansToDegrees << endl; 
  
  
  /***********************************************************************
   *
   * Need to check if the images have different resolutions (low or high)
   *
   ***********************************************************************/
  for(unsigned i = 1; i < wcsList.size(); i++){
    //cout << "L1068 i="<< i << "------------------" << endl;
    if (wcsList[0].data->binAxis != wcsList[i].data->binAxis){
      mixedBinning = true;      
    }  
    //cout << "L1074 mixedBinning=" << mixedBinning << endl;
    if(c3.imageType == 0){
      // Sky image
      //printf("L1073      xLeft0=%f xRight0=%f \n",  wcsList[0].data->xLeft * radiansToDegrees, 
      //     wcsList[0].data->xRight* radiansToDegrees);
      //printf("L1074      xLefti=%f xRighti=%f \n",  wcsList[i].data->xLeft* radiansToDegrees,
      //     wcsList[i].data->xRight* radiansToDegrees);
      
      double tempiRight = wcsList[i].data->xRight;
      double tempiLeft  = wcsList[i].data->xLeft;
      double temp0Right = wcsList[0].data->xRight;
      double temp0Left  = wcsList[0].data->xLeft;
      /* protect against possible errors in the case of having the 24h-division line within the image */
      if ((tempiRight* radiansToDegrees+180.0)<(temp0Right * radiansToDegrees)){
	tempiRight+=360.0*degreesToRadians;
      }
      if ((tempiLeft * radiansToDegrees+180.0)<(temp0Left * radiansToDegrees)){
	tempiLeft+=360.0*degreesToRadians;
      }
      if ((temp0Right* radiansToDegrees+180.0)<(tempiRight * radiansToDegrees)){
	temp0Right+=360.0*degreesToRadians;
      }
      if ((temp0Left * radiansToDegrees+180.0)<(tempiLeft * radiansToDegrees)){
	temp0Left+=360.0*degreesToRadians;
      }
      /*
      c3.data->xLeft = max(c3.data->xLeft, wcsList[i].data->xLeft);
      c3.data->xRight = min(c3.data->xRight, wcsList[i].data->xRight);
      */
      c3.data->xLeft = max(c3.data->xLeft, tempiLeft);
      c3.data->xRight = min(c3.data->xRight, tempiRight);
      //printf("L1078  c3.xLefti=%f c3.xRighti=%f \n",  c3.data->xLeft* radiansToDegrees,
      //     c3.data->xRight* radiansToDegrees);
      c3.data->yBottom = min(c3.data->yBottom, wcsList[i].data->yBottom);
      c3.data->yTop = max(c3.data->yTop,    wcsList[i].data->yTop);      
            
      //cout << i << ") crval1=" << setprecision(6) <<  setiosflags(ios::fixed) << wcsList[i].data->crval1 * radiansToDegrees << 
      //" crval2=" << setprecision(6) <<  setiosflags(ios::fixed)  << wcsList[i].data->crval2 * radiansToDegrees << 
    //. " xleft=" << setprecision(6) <<  setiosflags(ios::fixed)   << wcsList[i].data->xLeft * radiansToDegrees << 
      //" xright=" <<  setprecision(6) <<  setiosflags(ios::fixed)   << wcsList[i].data->xRight * radiansToDegrees << 
      //" ytop=" <<  setprecision(6) <<  setiosflags(ios::fixed)   << wcsList[i].data->yTop * radiansToDegrees << 
      //" ybottom=" <<  setprecision(6) <<  setiosflags(ios::fixed)  << wcsList[i].data->yBottom * radiansToDegrees << endl; 
      
    } else {
      printf("L1099 i=%d xLeft=%f xRight=%f \n",i, wcsList[0].data->xLeft,wcsList[i].data->xRight);
      c3.data->xLeft = min(c3.data->xLeft, wcsList[i].data->xLeft);
      c3.data->xRight = max(c3.data->xRight, wcsList[i].data->xRight);
      c3.data->yBottom = min(c3.data->yBottom, wcsList[i].data->yBottom);
      c3.data->yTop = max(c3.data->yTop, wcsList[i].data->yTop);
      c3.data->crpix1 = min(c3.data->crpix1, wcsList[i].data->crpix1);
      c3.data->crpix2 = min(c3.data->crpix2, wcsList[i].data->crpix2);
    }
  } // for(unsigned i = 1; i < wcsList.size()


  printf("L1109 c3 xLeft=%f xRight=%f \n", c3.data->xLeft* radiansToDegrees,c3.data->xRight* radiansToDegrees);    
  
  /*************************************************************************
   *
   * Find the number of pixels in the stacked image
   *
   **************************************************************************/    
  c3.data->setImageParameters();
  
  int nx3 = c3.data->nx;
  int ny3 = c3.data->ny;
  
  
  //cout << "L1122 nx3=" << nx3 << " ny3=" << ny3 << endl;
  
  /************************************************************
   *
   * Now populate the image
   *
   *************************************************************/
  
  // cout << "line 1034 nx3=" << nx3 << " ny3=" << ny3 << endl;
  stackedImage.resize( nx3 );
  
  /***************************************************************
   *
   *
   ****************************************************************/
  
  stackedExposureImage.resize( nx3 );
  stackedQualityImage.resize( nx3 );
  
  //cout << "L1140 nx3=" << nx3 << " ny3=" << ny3 << endl;
  for(int i = 0; i < nx3; i++)
    {
      stackedImage[i].resize( ny3 );
      stackedExposureImage[i].resize( ny3 );
      stackedQualityImage[i].resize( ny3 );
      
      for(int j = 0; j < ny3; j++)
        {
	  stackedImage[i][j] = -1.0;
          stackedExposureImage[i][j] = -1.0;
          stackedQualityImage[i][j] = -1;      
        }
    }
  
  /***********************************************************************
   *
   *
   * Loop through all the files
   *
   ***********************************************************************/
  
  /*
   * 2014-02-01 vny computing the max valued of frame time and dead fraction
   */
  //double deadFracMax=0.0;
  //double frameTimeMax=0.0;
  deadFracMax=0.0;
  frameTimeMax=0.0;
  //cout << "L1193 deadFrac=" << deadFrac << " frameTime=" << frameTime << endl;
  for(unsigned i = 0; i < fileList.size(); i++)
    {
      
      errstr << i << ") Adding " << fileList[i]  
	     << Sas::message(Msg::App, Msg::Sparse);
      
      DataSet *dataSet = dataSetServer->open( fileList[i], Dal::Read);
      
      int nx = 0, ny = 0;
      vector<vector<float> > image = loadFloatImage( dataSet, "PRIMARY", nx, ny);
      //cout << "L1171 .nx=" << nx  << " ny=" << ny << endl;
      //cout << "L1172 image.size1=" << image.size() << " size2=" << image[0].size() << endl;
      /*****************************************************************
       *
       * If high-resolution window and mix of low/high res windows divide by 4
       *
       *****************************************************************/ 
      //        if( mixedBinning && wcsList[i].data->binAxis == 1) // High resolution image
      //	{
      
      /**********************************************************
       *
       * 4 high-resolution pixels equal one low resolution pixel
       * Multiply image by 4 to account for this
       **********************************************************/
      //	    multiplyImage( image, 4.0);
      
      // } 
      
      /**********************************************************************
       *
       * Load the EXPOSURE image
       *
       **********************************************************************/
      vector<vector<float> > exposureImage;
      
      if(dataSet->hasBlock("EXPOSURE") ){
	//cout <<  "L1199 loading the exposure image from the input file \n ";
	exposureImage = loadFloatImage( dataSet, "EXPOSURE", nx, ny);
	//cout << "L1200 .nx=" << nx  << " ny=" << ny << endl;
	
      } else {
	
	float exposure = dataSet->attribute("EXPOSURE")->asReal();  // Exposure time
	exposureImage.resize(nx);
	//cout << "L1207 creating the exposure map from the EXPOSURE=" << exposure << endl;
	
	/******************************************************************
	 *
	 * Correct EXPOSURE for dead-fraction
	 * 
	 ********************************************************************/ 
	double constant = exposure * (1.0 - wcsList[i].data->deadFrac);

	//cout << "L1234 i=" << i << " deadFrac=" << wcsList[i].data->deadFrac 
	//     << " frameTime=" << wcsList[i].data->frameTime << endl;

	if (wcsList[i].data->frameTime > frameTimeMax){
	  frameTimeMax=wcsList[i].data->frameTime;
	}
	//if (wcsList[i].data->deadFrac > deadFracMax){
	  //deadFracMax=wcsList[i].data->deadFrac;
	//}
	// SPR-7299 - set the deadtime fraction from the maximum frame time.
	// this is now actually the MINIMUM deadtime fraction !
	if (frameTimeMax > 0.0) 
  	   deadFracMax=0.174/frameTimeMax;

	wcsList[i].data->deadFrac = 0.0;
	wcsList[i].data->frameTime = 0.0;

	//cout << "L1240 i=" << i << " deadFracMax=" << deadFracMax 
	//     << " frameTimeMax=" << frameTimeMax << endl;	
	//cout << "L1242 .nx=" << nx  << " ny=" << ny << endl;
	//cout << "L1243 image.size=" << image.size() << endl;
	
	for(int i1 = 0; i1 < nx; i1++)
	  {
	    exposureImage[i1].resize(ny);
	    for(int j = 0; j < ny; j++)
	      {
		if( !isNull(image[i1][j] ) )
		  exposureImage[i1][j] = constant;
		else
		  exposureImage[i1][j] = -1.0;
	      }
	  }
	
      }
      //cout << "L1271 i=" << i << " deadFracMax=" << deadFracMax 
      //   << " frameTimeMax=" << frameTimeMax << endl;	
      /*****************************************************************
       *
       * The QUALITY image
       *
       ******************************************************************/ 
      vector<vector<int16> >  qualityImage;
      /*****************************************************************
       *
       *
       * Need to check if QUALITY image is 8 (old) or 16 (new) bits
       *
       *****************************************************************/
      //cout << "L1244 .nx=" << nx  << " ny=" << ny << endl;
      //cout << "L1245 image.size=" << image.size() << endl;
      if(dataSet->hasBlock("QUALITY") )
	{
	  const Array* array = dataSet->array("QUALITY");
	  int type = array->dataType();
	    
	  if(type == Dal::Int16)
	    {
	      qualityImage = loadInt16Image( dataSet, "QUALITY", nx, ny);
	      
	    }
	  else
	    {
	      //cout << "L1256 resizeing qualityImage.nx=" << nx  << endl;
	      qualityImage.resize(nx);
	      vector<vector<int8> > int8QualityImage = loadInt8Image( dataSet, "QUALITY", nx, ny);
	      for(unsigned i1 = 0; i1 < int8QualityImage.size(); i1++)
		{
		  qualityImage[i1].resize(ny);
		  
		  for(unsigned j = 0; j < int8QualityImage[i1].size(); j++)
		    {
		      qualityImage[i1][j] = (int16)int8QualityImage[i1][j];
		      }
		}
		
	    }  
	}
      else
	{
	  qualityImage.resize(image.size() );
	  //cout << "L1273 resizeing image.size=" << image.size() << endl;
            for(unsigned i1 = 0; i1 < image.size(); i1++)
	      {
     	        qualityImage[i1].resize(image[i1].size());
                for(unsigned  j = 0; j < image[i1].size(); j++)
		  {
		    if( isNull(image[i1][j] ) )
		      qualityImage[i1][j] = -1;
                    else
		      qualityImage[i1][j] = 0;
		  } 
		
	      }  
	}
      
      dataSetServer->close(dataSet); // Finished with the file
	
        /***********************************************************************
	*
        * If a mixture of low/high resolution images, first convert any high-resolution
        * image to a low-resolution image before stacking
        *
        ************************************************************************/
      

	if( mixedBinning && wcsList[i].data->binAxis == 1) // High resolution image
	  {
	    
	    /**********************************************************************
	     *
	     * Convert to low resolution image (binaxis = 2)
	     *
	     ***********************************************************************/
	    //cout << "L1306 converting to low resolution nx=" << image.size() << " ny=" << image[0].size() << endl;
	    wcsList[i].data->convertToLowResImage( image, exposureImage, qualityImage);
	    
	  }
              

	//cout << i << ") Line 1314 adding image crval1=" <<  setprecision(10) << wcsList[i].data->crval1 * radiansToDegrees << 
	 //" crval2=" << setprecision(10) << wcsList[i].data->crval2 * radiansToDegrees << endl; 

	//cout << "L1318 nx=" << nx << " ny=" << ny << " image.size1=" << image.size() <<  " size2=" << image[0].size() << wcsList[i]<<endl;
	addImage(stackedImage, c3, image, wcsList[i]);
	//cout << "L1319 ----------------------   adding exposure image " << endl;
        addImage(stackedExposureImage, c3, exposureImage, wcsList[i]);
	//cout << "L1321 ----------------------   adding quality  image " << endl;
	addImage(stackedQualityImage, c3, qualityImage, wcsList[i]);
    } // for(unsigned i = 0; i < fileList.size() 


    /*******************************************************
    *
    * Interpolate across isolated null pixels
    *
    ******************************************************/
    errstr << "Interpolating across isolated null pixels  "  
	   << Sas::message(Msg::App, Msg::Sparse);
    
    refineImage( stackedExposureImage);
    refineQualityMap(stackedQualityImage, stackedImage);
    refineImage( stackedImage);
    
}


/*************************************************************************************
* This routine uses the image x and y coordinates of the stacked image
* plus the xOffset and yOffset of the image to interpolate
*
* This function is specialized for the quality image
*
*************************************************************************************/

void addImage(vector<vector<int16> > &stackedImage, 
              OMWCS &stackedWcs,
              vector<vector<int16> > &image, OMWCS &wcs)

{
 
    int nx = stackedImage.size();
    int ny = stackedImage[0].size();
    //cout << "L1360  nx=" << nx << " ny=" << ny << endl;
 
    Interpolation<int16> interpolator(image);

    for(int i = 0; i < nx; i++)
    {         
        double x1 = double(i + 1);
        for(int j = 0; j < ny; j++)
        {
      
	    double y1 = double(j + 1);
                                      
            /**************************************************************************
	    *
            *
            ***************************************************************************/
            double x2 = x1, y2 = y1;
            stackedWcs.data->convertX1Y1ToX2Y2(x1, y1, wcs, x2, y2);
    

	    if (wcs.data->withinImage(x2, y2) )
	    {   	  
	        int16 value = interpolator.BilinearInterpolation(x2, y2);                   
	        setPixel(stackedImage[i][j], value);
			 		                        
            }

	} // End of j loop
    } // End of i loop
}
/*************************************************************************************
*
* This routine uses the image x and y coordinates of the stacked image
* plus the xOffset and yOffset of the image to interpolate
*
*
*
*************************************************************************************/
void addImage(vector<vector<float> > &stackedImage, 
              OMWCS &stackedWcs,
              vector<vector<float> > &image, OMWCS &wcs)

{
 
    int nx = stackedImage.size();
    int ny = stackedImage[0].size();

    //cout << "L1407  nx=" << nx << " ny=" << ny << endl;

    /*********************************************************************************
    *
    * 
    *
    **********************************************************************************/
  
 
    Interpolation<double> interpolator(image);    
   
    for(int i = 0; i < nx; i++)
    {
        double x1 = double(i + 1);
                    
        for(int j = 0; j < ny; j++)
        {
	    double y1 = double(j + 1);
                                       
            /**************************************************************************
	    *
            *
            ***************************************************************************/
            double x2 = x1, y2 = y1;
            stackedWcs.data->convertX1Y1ToX2Y2(x1, y1, wcs, x2, y2);
		                
	    if (wcs.data->withinImage(x2, y2) )
	    {   

                float value = interpolator.BilinearInterpolation(x2, y2);            
                setPixel(stackedImage[i][j], value);
            			 		                        
            }

	} // End of j loop
    } // End of i loop
    /// RDS DEBUG ///
    double x2=1.0, y2=1.0;
    stackedWcs.data->convertX1Y1ToX2Y2(1.0,1.0,wcs, x2, y2);
           
}
/***************************************************
*
* Try to find a more accurate correlation coefficient
* from the grid of 2-d array of computed values
*
****************************************************/
void refineCorrelation(vector<vector<double> > &r, double dw, double &maxDx, 
                  double &maxDy, double &maxR)
{
    int nx = r.size();
    int ny = r[0].size();
    double r1[nx * ny];

    int n = 0;
    for (int i = 0; i < nx; i++)
        for (int j = 0; j < ny; j++)
	    r1[n++] = r[i][j];

    F90MODPROC(interpolate_mod, maximizer)(&nx, &ny, r1, &dw, &maxDx, &maxDy, &maxR);
    
}

/**********************************************************************
*
*
*
*
**********************************************************************/
void getMeanPhotometryConstants(vector<OMWCS> &stackedWcsList, OMWCS &wcs)
{
    double meanDeadFrac = 0.0, meanFrameTime = 0.0;

    for(unsigned i = 0; i < stackedWcsList.size(); i++)
    {
        meanDeadFrac += stackedWcsList[i].data->deadFrac;
        meanFrameTime += stackedWcsList[i].data->frameTime;

	//	cout << i << ") frametime=" << stackedWcsList[i].data->frameTime 
        //      << " exposure=" << stackedWcsList[i].data->exposure << endl;

    }

    wcs.data->deadFrac = meanDeadFrac / double(stackedWcsList.size() );
    wcs.data->frameTime = meanFrameTime / double(stackedWcsList.size() );
    wcs.data->filter = stackedWcsList[0].data->filter;
    wcs.data->deadFrac = 0.0;
    wcs.data->frameTime = 0.0;



}
/*******************************************************
*
*
******************************************************/

void setPixel( float &imagePixel, double value)
{

    if(isNull(value ) ) return;
 
    if( value < 0.0) return; 
    if(imagePixel < 0.0)
    {
        imagePixel = value;
    }
    else
    {
        imagePixel += value;
    } 
	
}	       
                      
/*******************************************************
*
*
*
*
******************************************************/

void setPixel( int16 &imagePixel, int16 value)
{

    if(isNull( value ) ) return;
 
    if( value < (int16)0) return; 

    if(imagePixel < (int16)0)
    {
        imagePixel = value;
    }
    else
    {
        imagePixel = max(imagePixel, value);
    } 

 	
}	       

/****************************************************************
*
*
*
*
*****************************************************************/
void createCorrelationFile(vector<vector<double> > &image, 
                           string file1, 
                           string file2, 
                           double corr,
                           double dx, double dy)
{

    string fileName = stringParameter("correlset");


    DataSet * set  = dataSetServer->open(fileName, Dal::Create);

  
  
    int nx = (int)image.size();
    int ny = (int)image[0].size();
  
    string name = "CORRELATION_00";
    addBlock(set, name, image);

    Block *block = set->block(name);

    *block->addAttribute("IMAGE1", "Reference image") = file1;
    *block->addAttribute("IMAGE2", "Aligned with reference image") = file2;
    *block->addAttribute("CORREL", "Cross-correlation coefficient") = corr;
    *block->addAttribute("XSHIFT", "Offset (pixels) between x origins") = dx;
    *block->addAttribute("YSHIFT", "Offset (pixels) between y origins") = dy;





    *set->addAttribute("WINDOWX0", "") = 0;
    *set->addAttribute("WINDOWDX", "") = nx;
    *set->addAttribute("WINDOWY0", "") = 0;
    *set->addAttribute("WINDOWDY", "") = ny;


    dataSetServer->close(set); // Finished with the file
       
} 
/****************************************************************
* 
*
*
*****************************************************************/
void updateCorrelationFile(vector<vector<double> > &image, 
                           string file1, 
                           string file2,
                           double corr,
                           double dx,
                           double dy)
{

    string fileName = stringParameter("correlset");

    //  cout << "Updating file " << fileName << endl;
    DataSet * set  = dataSetServer->open(fileName, Dal::Modify);

    int blockNum = set->blocks();  

    char name[30];
    sprintf(name, "CORRELATION_%02d", blockNum);

    addBlock(set, name, image);
 
    Block *block = set->block(name);

    *block->addAttribute("IMAGE1", "Reference image") = file1;
    *block->addAttribute("IMAGE2", "Aligned with reference image") = file2;


    *block->addAttribute("CORREL", "Cross-correlation coefficient") = corr;
    *block->addAttribute("XSHIFT", "Offset (pixels) between x origins") = dx;
    *block->addAttribute("YSHIFT", "Offset (pixels) between y origins") = dy;



    dataSetServer->close(set); // Finished with the file
       

}

/*************************************************************************************
*
*
*
* Returns 1 if the image is a sky image, 2 otherwise
* 
*
*************************************************************************************/
int getImageType( string file)
{


    int type = 1;
    DataSet *dataSet = dataSetServer->open( file, Dal::Read);
    
    if ( dataSet->hasAttribute("CDELT1") )
    {
        type = 0;
      
          
    }
    else
    {
        const Block *block = dataSet->block(0);
        if ( block->hasAttribute( "CDELT1" ) )
	{
	    type = 0;

	}
        
     
    }
    dataSetServer->close(dataSet); // Finished with the file
 
    return type;

}

/*****************************************************************************
* Find correlation peak(rImage, dw, maxDx, maxDy, maxR)
*
******************************************************************************/
void  findCorrelationPeak(vector< vector<double> > &image, int di, double dw, 
     double &maxDx, double &maxDy, double &maxR, int &numPeaks)
{


  //  cout << "FindCorrelationPeak- initial dx=" << maxDx << " dy=" << maxDy << " r=" << maxR << endl;


    int nx = (int)image.size();
    int ny = (int)image[0].size();

    double maxPeak = 0.0;    
    double maxSum = 0.0;


    vector<vector<bool> > used(image.size(), vector<bool>(image[0].size(), false) );

    numPeaks = 0;
    for(int i = 1; i < nx - 1; i++)
    {
             
        for(int j = 1; j < ny - 1; j++)
        {
	    if(used[i][j] ) continue;
            if(used[i][j - 1] ) continue;
            if(used[i][j + 1] ) continue;
            if(used[i - 1][j] ) continue;
            if(used[i + 1][j] ) continue;
            if(used[i - 1][j - 1] ) continue;
            if(used[i - 1][j + 1] ) continue;
            if(used[i + 1][j - 1] ) continue;
            if(used[i + 1][j + 1] ) continue;
       


            double f = (double)image[i][j];
	    //           if( f < maxPeak) continue;
	    if( image[i - 1][j] > f) continue;
	    if( image[i + 1][j] > f) continue;          
	    if( image[i][j - 1] > f) continue;
	    if( image[i][j + 1] > f) continue;
            if( image[i - 1][j - 1] > f) continue;
            if( image[i - 1][j + 1] > f) continue;
            if( image[i + 1][j - 1] > f) continue;
            if( image[i + 1][j + 1] > f) continue;


            double sum = (double)f + 
                         (double)image[i - 1][j] +
                         (double)image[i + 1][j] +
                         (double)image[i][j - 1] +
	                 (double)image[i][j + 1] +
                         (double)image[i - 1][j - 1] +
                         (double)image[i - 1][j + 1] +
                         (double)image[i + 1][j - 1] +
	                 (double)image[i + 1][j + 1];

             

            if(sum > maxSum)
	    {
                maxDx = double(i - di) * dw; 
                maxDy = double(j - di) * dw;  
                maxR = f;
		maxPeak = f;
                maxSum = sum;
		++numPeaks;
                used[i][j] = true;
                used[i][j - 1] = true;
                used[i][j + 1] = true;

                used[i - 1][j] = true;
                used[i + 1][j] = true;
                used[i - 1][j + 1] = true;
                used[i - 1][j - 1] = true;
                used[i + 1][j + 1] = true;
                used[i + 1][j - 1] = true;

		//		cout << "Peak found at dx=" << maxDx << " dy=" << maxDy << " R=" << f << endl;
            

            }
        }   

    }

    //  cout << "Number of peaks found=" << numPeaks << endl;
    // cout << "FindCorrelationPeak- final dx=" << maxDx << " dy=" << maxDy << " r=" << maxR << endl;
    // cout << "maxPeak=" << maxPeak << " maxSum=" << maxSum << endl; 

}
    

/*************************************************************
*
*
*************************************************************/
double totalCounts(vector<vector<float> > & image)
{

    double counts = 0.0;
    for(unsigned i = 0; i < image.size(); i++)
    {
        for(unsigned j = 0; j < image[i].size(); j++)
        {
	  if( !isNull(image[i][j] ) ) counts += (double)image[i][j];
        }
    }
    //   cout << "Total image counts=" << counts << endl;
    return counts;
    

}  
/************************************************************
*
*
*
************************************************************/
void setUsablePixels( ImageData &data )
{

    float nSigma1 = realParameter("nsigma");    
    int minNumPixels = intParameter("minnumpixels");

    data.x.clear();
    data.y.clear();
    data.f.clear();
    data.fSquared.clear();

 
    for(int k = 0; ;  k++)
    {
        float nSigma = nSigma1 - 0.5 * (float)k; 

        data.numberGoodPixels = 0; // Record total number of pixels above the background
        for(int i = 0; i < data.nx; i++)
        {
            double x = double(i + 1);
           for(int j = 0; j < data.ny; j++)
           {
               if( !isNull(data.image[i][j]) )
	       {
                   if ( data.image[i][j] >= 
                         ( data.backgroundImage[i][j] + max((float)2, nSigma * sqrt(data.backgroundImage[i][j])) ) )
	           {
	               double f1 = data.image[i][j] - data.backgroundImage[i][j];
                       if( f1 >= 0.0)
	               {
	                   double y = double(j + 1);
                           data.x.push_back(x);
                           data.y.push_back(y);
                           data.f.push_back( f1 );
                           data.fSquared.push_back(f1 * f1);
                           ++data.numberGoodPixels;

		       }
                   }
             
               }
           }

        }
        if( data.numberGoodPixels >= minNumPixels) break;
        if (nSigma <= 1.0) break;
    }

}
/************************************************************
*
*
************************************************************/
void removeBackground( ImageData &data )
{

    for(int i = 0; i < data.nx; i++)
    {  
        for(int j = 0; j < data.ny; j++)
        {
            if( !isNull(data.image[i][j]) )
	    {
               
                data.image[i][j] -= data.backgroundImage[i][j];  
            }
        }
    }
}
/****************************************************************
*
*
*
****************************************************************/
void setUpOffsets( int di, vector<int> &xOffset, vector<int> &yOffset)
{ 
 
    for(int i = 0;  i <= di; i++)
    {
        int dx = i;
        xOffset.push_back(dx);
        yOffset.push_back(0);
  
        
        for(int j = 1; j <= di; j++)       
	{
	    xOffset.push_back(dx);
            int dy = j;
            yOffset.push_back(dy);
            
            xOffset.push_back(dx);
            dy = -dy;
            yOffset.push_back(dy);
	}

        dx = -dx;  
        xOffset.push_back(dx);
        yOffset.push_back(0);
       
        for(int j = 1; j <= di; j++)       
	{
	    xOffset.push_back(dx);
            int dy = j;
            yOffset.push_back(dy);
            
            xOffset.push_back(dx);
            dy = -dy;
            yOffset.push_back(dy);
	}
    }      
}   
/***************************************************************************
*
*
* Attempts to align the list of images with respect to some reference image
*
*
* The algorithm
* EG 5 images
* The alignment list stores the following
*    index i index j   overlap fraction
*
*  Before sorting (First stage) (Done in initialise)
*
*1)  1)     1         2   0.1 
*2)  2)     1         3   0.1
*3)  3)     1         4   0.1
*4)  4)     1         5   0.5
*  
*5)  1)     2         3   0.1
*6)  2)     2         4   0.6
*7)  3)     2         5   0.1


*8)  1)     3         4   0.7
*9)  2)     3         5   0.6
*
*10) 1)     4         5   0.9
*
*
* After sorting into decreasing overlap fraction by Alignment class (Second stage)
*
* 1)        4         5   0.9
* 2)        3         4   0.7
* 3)        2         4   0.6
* 4)        3         5   0.6
* 5)        1         5   0.5
* 6)        1         2   0.1
* 7)        1         3   0.1
* 8)        1         4   0.1
* 9)        2         3   0.1
*10)        2         5   0.1
*
* Third stage - alignment (This routine)
*
* Set reference image to 4 - set isStacked[3] to true 
*
* 1) Attempt to align image 5 with image 4
*        If successful - set isStacked[4] to true
* 2) Attempt to align image 3 with image 4
*        If successful - set isStacked[2] to true
* 3) Attempt to align image 2 with image 4
*        If successful - set isStacked[1] to true
* 4) 
***************************************************************************/
void alignImages(vector<Alignment> &alignmentList,
                 vector<string> &fileList,
                 vector<OMWCS>  &wcsList,
                 vector<string> &stackedFileList, 
                 vector<OMWCS>  &stackedWcsList,
                 int &noStackedFiles)       
{


    if(alignmentList.size() == 0)
    {
        errstr << "No images can be aligned"
	       << Sas::warning("nofilestoalign");


        if(realParameter("mincorr") < 0.0) // Want to stack all the images, even those that haven't been aligned
        {

            for(unsigned i = 0; i < fileList.size(); i++)
            {
          
                stackedFileList.push_back(fileList[i] );  
                stackedWcsList.push_back(wcsList[i] );
             
         

	    }

        }
   

        noStackedFiles = (int)fileList.size();
       
        return;

    }
    vector<bool> isStacked(fileList.size(), false);

    //    int index = alignmentList[0]._iIndex;
    //    isStacked[index] = true;


    //    stackedFileList.push_back( fileList[index] );
    //    stackedWcsList.push_back(wcsList[index] );

    map<int,  map<int, int>  > alignmentMap;
 

    /*************************************************************
    *
    * Loop through the list of pairs of images (sorted into decreasing
    * order of overlap fraction)
    *
    ***************************************************************/
    bool firstStacked = true;
    for(int iter = 1;; iter++)
    {   
     
        bool anyStacked = false;

        for(unsigned k = 0; k < alignmentList.size(); k++)   
        {
	    if ( !alignmentList[k]._good) continue;
   
            
            unsigned i = alignmentList[k]._iIndex;
            unsigned j = alignmentList[k]._jIndex;

            if( isStacked[i] && isStacked[j] ) continue;   // Both images have been aligned so move on to the next
   
            if( !isStacked[i] && !isStacked[j] && !firstStacked) continue; // Neither image has been aligned so move on to the next
        
            if( !isStacked[i] && !firstStacked )  // i has not been stacked
	    {
	        /*****************************************************
	        * Swap indices
                *
                *****************************************************/
	        int k = i; 
                i = j;
                j = k;

	    }    
	    

            /****************************************************************
	    *
            * Load the image data for this file
            *
            ****************************************************************/ 
	    ImageData imageDatai;
	    
	    loadImageData(fileList[i], imageDatai);
             
  	    /*********************************************
	    *
	    * Try to stack image i with image j
            *
            **********************************************/
	    ImageData imageDataj;
   
            loadImageData(fileList[j], imageDataj);
	 	  
	    bool okay = false;
            double cr = 0.0, maxDx = 0.0, maxDy = 0.0;         
	    
	    process(fileList[i], wcsList[i], imageDatai, fileList[j], wcsList[j], 
                    imageDataj, okay, cr, noStackedFiles, maxDx, maxDy);
	 
	    if (okay ) // Cross-correlation succeeded
            {

	         anyStacked = true;
	        /*****************************************************************
	        *
	        * Record that image j has been successfully correlated with image i
                *
                ******************************************************************/
	        ++noStackedFiles;
                isStacked[i] = true;                    
	        isStacked[j] = true;                    
           

                alignmentMap[i][j] = k;   //
                if(firstStacked)
		{
                    stackedFileList.push_back(fileList[i] );  
                    stackedWcsList.push_back(wcsList[i] );
                }
  
                stackedFileList.push_back(fileList[j] );  
                stackedWcsList.push_back(wcsList[j] );

                /**********************************************************************
		* Record various parameters
                *
                **********************************************************************/
                alignmentList[k]._dx = maxDx;                
                alignmentList[k]._dy = maxDy;                
                alignmentList[k]._correlation = cr;                
                alignmentList[k]._good = true;
                firstStacked = false;
	

		// cout << "Successfully aligned " << fileList[j] << " with " << fileList[i] << endl;
                errstr << "Successfully aligned " << fileList[j] << " with " << fileList[i] << Sas::message(Msg::App, Msg::Sparse);
                errstr << setiosflags(ios::fixed) << "Correlation coefficient=" << setprecision(2) << cr 
                       << " dx=" << setprecision(2) << maxDx << " dy=" << setprecision(2) << maxDy << Sas::message(Msg::App, Msg::Sparse);

             
            }
            else
	    {	  
	        alignmentList[k]._good = false;

            }

        }
           
        /**********************************************************************
	*
	* If we have stacked all the files then quit
        * If none have been stacked during this iteration break
        **********************************************************************/

        if( !anyStacked) break;
        if( stackedFileList.size() == fileList.size() ) break;
	  
    }

    /**********************************************************************
    *
    * May want to align all the images- even those that couldn't be aligned
    *
    ************************************************************************/
    if(realParameter("mincorr") < 0.0) // Want to stack all the images, even those that haven't been aligned
    {

        for(unsigned i = 0; i < fileList.size(); i++)
        {
	  if( !isStacked[i] )
	  {
          
                stackedFileList.push_back(fileList[i] );  
                stackedWcsList.push_back(wcsList[i] );
                      

	  }
        }
        noStackedFiles = (int)fileList.size();
    }    

    if(stackedFileList.size() < 1)
    {
        errstr << "No images have been successfully aligned"
	       << Sas::warning("nofilestostack");

        exit(1);

    }

    errstr << "Finished cross-correlating- number of images successfully aligned=" 
           << noStackedFiles << Sas::message(Msg::App, Msg::Sparse);

    /*************************************************************
    *
    * Have now finished aligning
    *
    *
    *************************************************************/
    map<int, map<int, int> >::iterator iter; 

    double minDx = 100000.0;
    double maxDx = -minDx;
    double minDy = 100000.0;
    double maxDy = -minDy;

    double sum1 = 0.0, sum2 = 0.0;
    int count = 0;

    for(iter = alignmentMap.begin(); iter != alignmentMap.end(); ++iter)
    {
        int i = iter->first;

        errstr << fileList[i] << " successfully aligned with the following images"
               << Sas::message(Msg::App, Msg::Sparse);

        map<int, int> list = iter->second;
        map<int, int>::iterator  iter1;
        i = 0;
        for(iter1 = list.begin(); iter1 != list.end(); ++iter1, ++i) 
        {
	  int j = iter1->first;
          int k = iter1->second;            
	  errstr << i << ") " << fileList[j] << setiosflags(ios::fixed)
                              << " R=" << setprecision(2) << alignmentList[k]._correlation
                              << " dx=" << setprecision(2) << alignmentList[k]._dx
                              << " dy=" << setprecision(2) << alignmentList[k]._dy
                      << Sas::message(Msg::App, Msg::Sparse);

          ++count;
          sum1 += alignmentList[k]._dx;
          sum2 += alignmentList[k]._dy;
          minDx = min(minDx, alignmentList[k]._dx);
          maxDx = max(maxDx, alignmentList[k]._dx);
          minDy = min(minDy, alignmentList[k]._dy);
          maxDy = max(maxDy, alignmentList[k]._dy);

        }  
    }


    if(count > 1)
    {
        double meanDx = sum1 / (double)count;
        double meanDy = sum2 / (double)count;
        errstr << "Mean dx=" << setiosflags(ios::fixed) << setprecision(2) <<  meanDx << " mean dy=" << meanDy 
               << " min dx=" << minDx << " max=" << maxDx << " min dy=" << minDy 
               << " max=" << maxDy << " (n=" << count << ")"
               << Sas::message(Msg::App, Msg::Sparse);

    }
}

/****************************************************************************
*
*
*
*****************************************************************************/
void refineImage(vector<vector<float> >  &image)
{
 
    vector<vector<float> > tempImage = image;

    int nx = (int)image.size() - 1;
    int ny = (int)image[0].size() - 1;


    for(int i = 0; i <= nx; i++)
    {
        for(int j = 0; j <= ny; j++)
        {
	    if( isNull(image[i][j] ) || image[i][j] < 0.0 )
	    {

	        fixPixel(image, i, j, nx, ny, tempImage[i][j]);

            }
        }
    }
    image = tempImage;
}
/****************************************************************************
*
*
*****************************************************************************/
void fixPixel(vector<vector<float> >  &image, int i, int j, 
                int nx, int ny, float &tempImage)
{
    int count = 0;
    double sum = 0.0;
    int dx = 1;
    int il = max( 0, i - dx);
    int iu = min(nx, i + dx);

    int jl = max( 0, j - dx);
    int ju = min(ny, j + dx);

    for(int i1 = il; i1 <= iu; i1++)
    {

        for(int j1 = jl; j1 <= ju; j1++)
        {
	    if( i1 == i && j1 == j) continue;
            if (!isNull(image[i1][j1]) && image[i1][j1] >= 0.0 )
	    {
                ++count;
                sum += (double)image[i1][j1];
            } 
        }

    }

    if(count > 1) 
    {
        tempImage = sum / (double)count;
	
    }
}

/****************************************************************************
*
*
*
*
*****************************************************************************/
void refineQualityMap(vector< vector<int16> > &qualityImage, vector< vector<float> > &image)
{


    int nx = (int)image.size();
    int ny = (int)image[0].size();

    const int maxCount = 1; 

    for(int i = 0; i < nx; i++)
    {     
        for(int j = 0; j < ny; j++)
	{
	    if(image[i][j] >= 0.0)
	    {
	        if(qualityImage[i][j] < 1)
                    qualityImage[i][j] = 0;
            }              
	    //	    else
	    // {
            //    if(qualityImage[i][j] < 2)
            //        qualityImage[i][j] = 1;              
	    // }  
                     

        }   

    }

    for(int i = 0; i < nx; i++)
    {
        int count = 0;
        for(int j = 0; j < ny; j++)
	{
	    if( image[i][j] >= 0.0)
	    {
                qualityImage[i][j] = 1;              
	        if(++count > maxCount) break;
            }         

        }   

    }


    for(int i = nx - 1; i >= 0; i--)
    {
        int count = 0;
        for(int j = 0; j < ny; j++)
	{
	    if( image[i][j] >= 0.0)
	    {
                qualityImage[i][j] = 1;              
	        if(++count > maxCount) break;
            }         


        }   


    }

    for(int j = 0; j < ny; j++)
    {
        int count = 0;
        for(int i = 0; i < nx; i++)
	{
	    if( image[i][j] >= 0.0)
	    {
                qualityImage[i][j] = 1;              
	        if(++count > maxCount) break;
            }         
        }   
    }


    for(int j = ny - 1; j >= 0; j--)
    {
        int count = 0;
        for(int i = 0; i < nx; i++)
	{
	    if( image[i][j] >= 0.0)
	    {
                qualityImage[i][j] = 1;              
	        if(++count > maxCount) break;
            }         

        }   

    }

}
/*****************************************************************
*
* If high-resolution window divide by 4
*
****************************************************************/ 
void multiplyImage( vector< vector<float> > &image, const float constant)
{

    int nx = (int)image.size();
    int ny = (int)image[0].size();

    for(int i = 0; i < nx; i++)
    {
        for(int j = 0; j < ny; j++)
        {
	    if( !isNull(image[i][j] ) && image[i][j] > 0.0 )
	    {
	         image[i][j] *= constant;
		                 
	    }
        }

    }
}  
/*********************************************************************
*
* Reads WCS data from each input file
*
* There may be image files from more than one observation
*
*
*********************************************************************/
void initialise(map<string, vector<string> >     &fileListMap, 
                map<string, vector<OMWCS> >      &wcsListMap, 
                map<string, vector<Alignment> >  &alignmentListMap)
{


    errstr << "Initialising" << Sas::message(Msg::App, Msg::Sparse);


    loadData(fileListMap, wcsListMap); 


    vector<Alignment> alignmentList;



    /*********************************************************************************
    *
    * If sky images then need to check the astrometry corrections
    *
    **********************************************************************************/
 

    map<string, vector<string> >::iterator iter; 
    for(iter = fileListMap.begin(); iter != fileListMap.end(); ++iter)
    {
        string obsId = iter->first;

        if( imageType == 0)
	{
	  //   cout << "Checking astrometry for obsid " << obsId << endl;
            OMWCS wc3( wcsListMap[obsId][0] ); // The parameters for the mosaiced image


            checkAstrometry( wcsListMap[obsId], wc3 );
            updateImageHeaders( fileListMap[obsId], wcsListMap[obsId]);
        }
        setAlignmentParameters( fileListMap[obsId], wcsListMap[obsId], 
                                    alignmentListMap[obsId] ); 
    
    }


}

/************************************************************************
*
*
*
*************************************************************************/
void setAlignmentParameters( vector<string>       &fileList, 
                             vector<OMWCS>        &wcsList, 
                             vector<Alignment>    &alignment)
          
{

    alignment.resize(0);
    int size = (int)fileList.size();
 
    double minFraction = realParameter( "minfraction" ); // The minimum area fraction
    int minNumPixels = intParameter("minnumpixels");


   

    vector<int> numberGoodPixels;
    for(int i = 0; i < size; i++)
    {

        ImageData data;
        loadImageData(fileList[i], data);
     
	setUsablePixels(data);
  
        numberGoodPixels.push_back( data.numberGoodPixels);    
    }
   
    for(int i = 0; i < size; i++)
    {
        
        if( numberGoodPixels[i] < minNumPixels)
	{
	    continue; 
        }

        for(int j = i + 1; j < size; j++)
	{
         
            if(numberGoodPixels[j] < minNumPixels)
	    {
	        continue; 
            }
            double fraction = wcsList[i].data->overlapFraction( wcsList[j] );
            if(fraction < minFraction) continue;         
            alignment.push_back( Alignment(i, j, fraction, numberGoodPixels[i], numberGoodPixels[j]) );
           
        }
    }



    if(alignment.size() < 2) return;
    /*****************************************************************
    *
    *
    *****************************************************************/

    std::sort(alignment.begin(), alignment.end() );





    //    return;

    // for(unsigned k = 0; k < alignment.size(); k++)
    // {
    //  int i = alignment[k]._iIndex;
    //  int j = alignment[k]._jIndex;

    //      cout << k << ") " << fileList[i] << " " << fileList[j] << " " << alignment[k]._overlapFraction << endl;
    //  }    

    // exit(0);
}

/*******************************************************************
*
*
*
*
******************************************************************/
void loadData(map<string, vector<string> > &fileListMap,
              map<string, vector<OMWCS> >  &wcsListMap) 

{




    /********************************************************
    *
    * Get the list of images to be stacked from the parameter line
    *
    *********************************************************/
    vector<string> inputFileList = stringListParameter("imagesets");

    if( inputFileList.size() == 0)
    {


 
        Sas::fatal("NoFiles", "No input files given on the command-line");



    }
    


    /******************************************************
    *
    * Get the image type (0=sky, 1= normal)
    *
    ****************************************************/
    imageType = getImageType(inputFileList[0]);
 

    /********************************************************
    *
    * Undistort the images if unrotated
    *
    *********************************************************/

    //    if(imageType == 1) 
    // {
    //    undistortImages();
    //    inputFileList = stringListParameter("imagesets");
    // }






    /*****************************************************
    *
    * Check that the files exist
    *
    *****************************************************/
    for(unsigned i = 0; i < inputFileList.size(); i++)
    {

      //      cout << i << ") line 2445 loading data from " << inputFileList[i] << endl;
        if ( !dataSetServer->exists(inputFileList[i] ))
	{
            Sas::fatal("FileDoesNotExist-", inputFileList[i]);
   
        }

        /**************************************************
	*
        * Check for identical files
        *
        ***************************************************/

        for(unsigned  j = 0; j < i; j++)
	{
	    if (inputFileList[i] == inputFileList[j] ) 
	    {
                Sas::warning("Identical input files-", inputFileList[i]);
 
            }

        }

    }
    
    vector<OMWCS> inputWcsList( inputFileList.size(), OMWCS(imageType) );



    /******************************************************************
    *
    * Load WCS data from each file and check that the files are of
    * the same type (sky or normal images)
    *
    *******************************************************************/
    loadWcsData(inputFileList[0], inputWcsList[0] ); // Compare each image type with this one




    //	    cout << "0) Line 2530 crval1=" << setprecision(6) <<  setiosflags(ios::fixed) << inputWcsList[0].data->crval1 * radiansToDegrees << 
    //	                   " crval2=" << setprecision(6) <<  setiosflags(ios::fixed)  << inputWcsList[0].data->crval2 * radiansToDegrees << 
    //                      " xleft=" << setprecision(6) <<  setiosflags(ios::fixed)   << inputWcsList[0].data->xLeft * radiansToDegrees << 
    //                       " xright=" <<  setprecision(6) <<  setiosflags(ios::fixed)   << inputWcsList[0].data->xRight * radiansToDegrees << 
    //                       " ytop=" <<  setprecision(6) <<  setiosflags(ios::fixed)   << inputWcsList[0].data->yTop * radiansToDegrees<< 
    //	                   " ybottom=" <<  setprecision(6) <<  setiosflags(ios::fixed)  << inputWcsList[0].data->yBottom * radiansToDegrees << endl; 

   










    /****************************************************************
    *
    * Keep data for each observation separate
    *
    *****************************************************************/


    string obsId = inputWcsList[0].data->_obsId;
    fileListMap[obsId].push_back( inputFileList[0] );
    wcsListMap[obsId].push_back( inputWcsList[0] );



    for(unsigned i = 1; i < inputFileList.size(); i++)
    {



        /*********************************************************
	*
        * Check the image type
        *
        *********************************************************/	
	if( getImageType( inputFileList[i]) != imageType)
        {

            Sas::fatal("Different image types-", "Cannot process images of different types!");

    
	}
        
        loadWcsData(inputFileList[i], inputWcsList[i] ); 



	//	cout << i << ") Line 2583 crval1=" << setprecision(6) <<  setiosflags(ios::fixed) << inputWcsList[i].data->crval1 * radiansToDegrees << 
	//                  " crval2=" << setprecision(6) <<  setiosflags(ios::fixed)  << inputWcsList[i].data->crval2 * radiansToDegrees << 
        //                   " xleft=" << setprecision(6) <<  setiosflags(ios::fixed)   << inputWcsList[i].data->xLeft * radiansToDegrees << 
        //                   " xright=" <<  setprecision(6) <<  setiosflags(ios::fixed)   << inputWcsList[i].data->xRight * radiansToDegrees << 
        //                   " ytop=" <<  setprecision(6) <<  setiosflags(ios::fixed)   << inputWcsList[i].data->yTop * radiansToDegrees << 
	//                   " ybottom=" <<  setprecision(6) <<  setiosflags(ios::fixed)  << inputWcsList[i].data->yBottom * radiansToDegrees << endl; 

   



        string obsId = inputWcsList[i].data->_obsId;
        fileListMap[obsId].push_back( inputFileList[i] );
        wcsListMap[obsId].push_back( inputWcsList[i] );


    } // End of loading WCS data from images
      
}
/*****************************************************************************
*
*
* NB raOffset was divided by cos(decPnt)
*
*
******************************************************************************/
void checkAstrometry( vector<OMWCS> &wcsList, OMWCS &wc3 )
{

  //   return; 
    if(wcsList.size() < 1) 
    {
    
        return;
    }

    vector<OMWCS> timeList(imageType);

    for(unsigned i = 0; i < wcsList.size(); i++)
    {
        if( wcsList[i].data->_astrometry._poscorrOkay)
        {
	    timeList.push_back( wcsList[i] );     // NB CEVAL1 and CRVAL2 will haqve been corrected by omatt

        }      
    }


    if(timeList.size() < 1) 
    {
        errstr << "No sky images have had a successful astrometry correction"
	       << Sas::warning("noAstrometryCorrection");
        return;

    }

 
    /*************************************************************
    *
    * Sort into increasing time order
    *
    **************************************************************/
    std::sort(timeList.begin(), timeList.end() );


    errstr << "The sorted list of astrometry corrections\n" << Sas::message(Msg::App, Msg::Sparse);

    errstr << setw(18) << right << "Time (MJD)" << setw(12) << right << "RA offset" << right << setw(14) 
	   << "Dec offset" << Sas::message(Msg::App, Msg::Sparse);



    vector<double> raOffset, decOffset;



    for(unsigned i = 0; i < timeList.size(); i++)
    {
        errstr.unsetf(ios::fixed);
	  
        errstr << noshowpos << i << ") "
	       << setw(15) << setprecision(6) <<  setiosflags(ios::fixed) << timeList[i].data->_mjdMid  
               << setw(12) << setprecision(2)  << setiosflags(ios::fixed) << showpos << showpoint 
	       << timeList[i].data->_astrometry._raOffset 
               << setw(14) << setprecision(2)  << setiosflags(ios::fixed) << showpos << showpoint 
	       << timeList[i].data->_astrometry._decOffset 
               << Sas::message(Msg::App, Msg::Sparse);

        raOffset.push_back( timeList[i].data->_astrometry._raOffset ); 
        decOffset.push_back( timeList[i].data->_astrometry._decOffset ); 




    }
    errstr << noshowpos;    

    /************************************************************************
    *
    *
    *
    ************************************************************************/

    if( raOffset.size() > 3)
    {
        MadVal<double>  mad1(raOffset);

        double nSigma = 3.0;
          
        mad1.test( nSigma );
        mad1.listBadValues("MAD test- Possible bad astrometry RA offsets");

        MadVal<double>  mad2(decOffset);
             
        mad2.test( nSigma );
   
        mad2.listBadValues("MAD test- Possible bad astrometry Dec offsets");

	
    }

    /****************************************************************************
    *
    *  For those images with no astrometry correction compute an interpolated value
    *
    *
    ****************************************************************************/
    int n = (int)timeList.size();
   
    double startTime = timeList[0].data->_mjdMid;

    double endTime = timeList[n - 1].data->_mjdMid;


    double meanRaOffset = 0.0, meanDecOffset = 0.0;




    for(unsigned i = 0; i < wcsList.size(); i++)
    {

 
        /******************************************************************************
	*
	* If an astrometry correction was done CRVAL1 and CRVAL2 will have been corrected
        *
        *****************************************************************************/
      //cout << "L2803 i="<< i << endl;

        if( wcsList[i].data->_astrometry._poscorrOkay) // An astrometry correction was done, correct CRVAL1 and CRVAL2
	{
            wcsList[i].data->_astrometry._posintrp = false;
          
        }
        else   // An astrometry correction was done, correct CRVAL1 and CRVAL2
        {

	  //	    cout << "Computing ra/dec shifts for file " << i << ") time=" <<  wcsList[i].data->_mjdMid << endl;

	    double dx = 0.0, dy = 0.0;
	    double time = wcsList[i].data->_mjdMid;
            if(time <= startTime)             // Time before the first correction time
	    {
	        dx =  timeList[0].data->_astrometry._raOffset;
                dy =  timeList[0].data->_astrometry._decOffset;
 
            }

            else if(time >= endTime)          // Time after the last correction time   
	    {

               
              dx = timeList[n - 1].data->_astrometry._raOffset;
              dy = timeList[n - 1].data->_astrometry._decOffset;
 

            }
            else  // Interpolate corrections
	    {

	        double time1 = -1.0, time2 = -1.0;
                double dx1 = 0.0, dx2 = 0.0;
                double dy1 = 0.0, dy2 = 0.0;
                


                /*******************************************************
		*
		* Find the first time less than the observation time
                *
                *******************************************************/
 
	        for(int j = n - 1; j >=0; j--)
		{
                    double t = timeList[j].data->_mjdMid;
                    
  
    
		    if(t <= time)   
		    { 
                   
		        time1 = t;
                        dx1 = timeList[j].data->_astrometry._raOffset;
                        dy1 = timeList[j].data->_astrometry._decOffset;
                        break;   // Found the time
   
                    }  
                }
                     
                /*******************************************************
		*
		* Find the first time greater than the observation time
                *
                *******************************************************/
 

	        for(int j = 0; j < n; j++)
		{
                    double t = timeList[j].data->_mjdMid;
                    
  
		    if(t >= time)
		    { 
                   
		        time2 = t;
                        dx2 = timeList[j].data->_astrometry._raOffset;
                        dy2 = timeList[j].data->_astrometry._decOffset;
                        break;
                    }  
                }

                /**************************************************************
		*
		* Interpolate
                *
		***************************************************************/
                double dt = time2 - time1;
                if( dt < 1e-6)
		{
		  dx = dx1;
                  dy = dy1; 

                } 
                else
		{                                
                    double deltaX = dx2 - dx1;
                    double deltaY = dy2 - dy1;

                    double deltaT = time - time1;
                    dx = dx1 + deltaX * deltaT / dt;  
                    dy = dy1 + deltaY * deltaT / dt;
                }                        		   
           

            }
	    //  cout << "omstack Line 2716 Interpolated dx=" << dx << "  " << "dy=" << dy << endl; 



            /*****************************************************************
	    *
            * Correct CRVAL1 and CRVAL2
	    *
            *******************************************************************/

            wcsList[i].data->_astrometry._posintrp = true;
            wcsList[i].data->_astrometry._raOffset = dx;
            wcsList[i].data->_astrometry._decOffset = dy;



	    //cout << i << ") crpix1=" <<  wcsList[i].data->crpix1 <<  " uncorrected crval1=" << setprecision(10) << wcsList[i].data->crval1 * radiansToDegrees << endl;
	    //cout << i << ") crpix2=" <<  wcsList[i].data->crpix2 <<  " uncorrected crval2=" << setprecision(10) << wcsList[i].data->crval2 * radiansToDegrees << endl;

       


            wcsList[i].data->crval1 += (dx / 3600.0) * degreesToRadians;
            wcsList[i].data->crval2 += (dy / 3600.0) * degreesToRadians;



	    /****************************************************************
	    *
            *
            *
            *****************************************************************/
            wcsList[i].data->_cdecz = cos( wcsList[i].data->crval2 );
            wcsList[i].data->_sdecz = sin( wcsList[i].data->crval2 );
            wcsList[i].data->getCornerCoords();


	    // cout << i << ") crpix1=" <<  wcsList[i].data->crpix1 <<  " corrected crval1=" << setprecision(10) << wcsList[i].data->crval1 * radiansToDegrees << endl;
	    // cout << i << ") crpix2=" <<  wcsList[i].data->crpix2 <<  " corrected crval2=" << setprecision(10) << wcsList[i].data->crval2 * radiansToDegrees << endl;


        } // End of poscorrokay 

        meanRaOffset += wcsList[i].data->_astrometry._raOffset;
        meanDecOffset += wcsList[i].data->_astrometry._decOffset;
      
    }


    /*******************************************************************
    *
    * The mean offsets of all the images
    *
    ********************************************************************/
    wc3.data->_astrometry._poscorrOkay = true;

    double cs = cos(wc3.data->_decPnt * degreesToRadians);

    wc3.data->_astrometry._raOffset = cs * meanRaOffset / (double)wcsList.size();
    wc3.data->_astrometry._decOffset = meanDecOffset / (double)wcsList.size();

} 

/**************************************************************************
*
* 
*
*
*
***************************************************************************/           

void updateImageHeaders( vector<string> &fileList, vector<OMWCS> &wcsList )
{

    for(unsigned i = 0; i < wcsList.size(); i++)
    {
        DataSet *dataSet = dataSetServer->open( fileList[i], Dal::Modify);
        if( wcsList[i].data->_astrometry._poscorrOkay)   // No need to interpolate 
	{
            *dataSet->addAttribute("POSINTRP", "") = false;
        }
        else
	{
            *dataSet->addAttribute("POSINTRP", "") = true;
        }
    
	*dataSet->addAttribute("RA_OFF", "Astrometric RA correction", "arcsec") = 
                     cos(wcsList[i].data->_raPnt * degreesToRadians) * wcsList[i].data->_astrometry._raOffset;
	*dataSet->addAttribute("DEC_OFF", "Astrometric Dec correction", "arcsec") = wcsList[i].data->_astrometry._decOffset;

        dataSetServer->close(dataSet); // Finished with the file
 
    }

}
/**********************************************************************************
*
*
*
************************************************************************************/
void undistortImages()
{

    /********************************************************
    *
    * Get the list of images to be stacked from the parameter line
    *
    *********************************************************/
    vector<string> fileList = stringListParameter("imagesets");
 
    vector<string> tempList(fileList.size() );
    for(unsigned i = 0; i < fileList.size(); i++)
    {
        createFileName( tempList[i] );

        //cout << i << ") original file name=" << fileList[i] << " new=" << tempList[i] << endl;

        string command = "omundistortimage set=" + fileList[i] + " outset=" + tempList[i];

        int k = system( command.c_str() );
        if (k != 0)
	{
            cout << "Exit value =" << k << endl;
        }
    }
 
    /***************************************************************
    *
    * Set the parameter "imagesets" to these temporary files
    *
    ************************************************************/ 
    setParameter("imagesets", tempList);
}
/*************************************************************
*
*
*
**************************************************************/
void createFileName( string & tempFileName )
{

    for(int i = 0; i < 1000; i++)
    {
        char name[50];
        sprintf(name, "OMSTACKTEMP%03d.FITS", i);
        if ( !dataSetServer->exists(name) )
	{
	  tempFileName = name;
          return;
	}
    }
}


/*************************************************************
*
*
*
*
**************************************************************/
void removeTemporaryFiles()
{
 
    vector<string> fileList = stringListParameter("imagesets");
  
    for(unsigned i = 0; i < fileList.size(); i++)
    {
        if (remove( fileList[i].c_str() ) == -1)
	{



        }
    }

}










