C ESA (C) 2000-2018 
C
C This file is part of ESA's XMM-Newton Scientific Analysis System (SAS).
C
C    SAS is free software: you can redistribute it and/or modify
C    it under the terms of the GNU General Public License as published by
C    the Free Software Foundation, either version 3 of the License, or
C    (at your option) any later version.
C
C    SAS is distributed in the hope that it will be useful,
C    but WITHOUT ANY WARRANTY; without even the implied warranty of
C    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
C    GNU General Public License for more details.
C
C    You should have received a copy of the GNU General Public License
C    along with SAS.  If not, see <http://www.gnu.org/licenses/>.
      program fellipse
      implicit none
      save

! ..tests the ellipse fitting routine

! ..declare
      integer          i,j


! ..for generating and storing the noisy ellipse data
      integer          n,nmax
      parameter        (nmax=500)
      double precision x(nmax),y(nmax),x0,y0,rmaj,rmin,tilt,   &
                      co,si,rr1,rr2,xx,yy,angmin,angmax,ampl,  &
                      dangle,angle


! ..for the results of fitting an ellipse 
      double precision fit_xcenter,fit_ycenter,ttilt,                 &
                      xfoci1,yfoci1,xfoci2,yfoci2,                    & 
                      semi_major,semi_minor,foci_dist,eccen,    &
                      a,asig,b,bsig,c,csig,d,dsig,e,esig,f,fsig,      &
                      chisq


! ..for the random number generator
      integer          iseed
      double precision rvec(3*nmax)
     


! ...formats
 01   format(1x,a,1p10e12.4)
 02   format(1x,a,1pe11.3,' +/-',1pe11.3)
 03   format(1x,1pe11.3,' * x**2 +'/, &
            1x,1pe11.3,' * y**2 +',/, &
            1x,1pe11.3,' * x*y  +',/, &
            1x,1pe11.3,' * x    +',/, &
            1x,1pe11.3,' * y    +',/, &
            1x,1pe11.3,'        = 0')


      write(6,*) ' '
      write(6,*) 'lets generate of a noisy ellipse data set:'
      write(6,*) 'give negative integer random number seed => '
      read(5,*)   iseed
      iseed = -abs(iseed)

 05   write(6,*) ' '
      write(6,*) 'give number of data points (less than 500) =>'
      read(5,*), n


      write(6,*) ' '
      write(6,*) 'give the center of the ellipse =>'
      read(5,*)  x0,y0


      write(6,*) ' '
      write(6,*) 'give lengths of the major and minor axes =>'
      read(5,*)  rmaj,rmin

      write(6,*) ' '
      write(6,*) 'give the angle between the major axis and the x-axis,'
      write(6,*) 'the so called tilt of the ellipse (in degrees) =>'
      read(5,*)  tilt

 06   write(6,*) ' '
      write(6,*) 'give the minimum and maximum angles (in degrees) '
      write(6,*) 'to generate the noisy data between. for example, '
      write(6,*) '0 360 generates data over the entire x-y plane,'
      write(6,*) '0 90 generates data only in the first quadrant,'
      write(6,*) '45 135 generates data between 45 and 135 degrees =>'
      read(5,*)  angmin,angmax
      if (angmin .eq. angmax) goto 06

      write(6,*) ' '
      write(6,*) 'give the amplitude of the noise (zero = no noise) =>'
      read(5,*)  ampl



c..get the random numbers
       call ranmar(iseed,rvec,3*n)


c..generate the noisy inclined ellipse
      co     = cosd(tilt)
      si     = sind(tilt)
      dangle = angmax - angmin
      j = 1
      do i = 1, n
       angle = angmin + dangle*rvec(j)
       rr1   = rmaj + ampl*(1.0d0 - 2.0d0 * rvec(j+1))
       rr2   = rmin + ampl*(1.0d0 - 2.0d0 * rvec(j+2))
       xx    = rr1*cosd(angle)
       yy    = rr2*sind(angle)
       x(i)  = co*xx - si*yy + x0
       y(i)  = si*xx + co*yy + y0
       j     = j + 3
      enddo


c..find the best fit
      call fitell(x,y,n,    &
                 fit_xcenter,fit_ycenter,xfoci1,yfoci1,xfoci2,yfoci2, &
                 semi_major,semi_minor,foci_dist,ttilt,eccen,         &
                 a,asig,b,bsig,c,csig,d,dsig,e,esig,f,fsig,chisq)


c..some results of the fit
      write(6,*) ' '
      write(6,*)  'results of fitting the data to an ellipse:'
      write(6,01) 'center at  :',fit_xcenter,fit_ycenter
      write(6,01) 'foci1 at   :',xfoci1,yfoci1
      write(6,01) 'foci2 at   :',xfoci2,yfoci2
      write(6,01) 'smajor axis:',semi_major
      write(6,01) 'sminor axis:',semi_minor
      write(6,01) 'foci_dist  :',foci_dist
      write(6,01) 'tilt       :',ttilt
      write(6,01) 'eccentric  :',eccen


      write(6,*) ' '
      write(6,01) 'equation of the best fit ellipse:'
      write(6,03) a,b,c,d,e,f

      write(6,*) ' '
      write(6,01) 'uncertainties in the coefficients:'
      write(6,02) 'a=',a,asig,'b=',b,bsig,'c=',c,csig,  &
                 'd=',d,dsig,'e=',e,esig,'f=',f,fsig
      write(6,01) 'chi squared:',chisq




c..write out the random points for plots
c      do i = 1, n
c       write(6,*) x(i), y(i)
c      enddo
c
c..write out the fitted ellipse for plots
c      co = cosd(ttilt)
c      si = sind(ttilt)
c      do i=1,361
c       angle = 1.0d0 + (i-1) * 360.0d0
c       xx    = semi_major*cosd(angle) 
c       yy    = semi_minor*sind(angle) 
c       xxx   = co*xx - si*yy + fit_xcenter
c       xxx   = si*xx + co*yy + y0
c       write(2,*) xxx, yyy
c      enddo


      goto 05
      end






      subroutine fitell(x,y,n,  &
                       x0,y0,xfoci1,yfoci1,xfoci2,yfoci2, &
                       rmaj,rmin,foci_dist,tilt,eccen,    &
                       a,asig,b,bsig,c,csig,d,dsig,e,esig,f,fsig,chisq)
      implicit none
      save

c..this routine fits data to an ellipse.

c..input:
c..x(1:n) = x-coordinates of the data
c..y(1:n) = y coordinates of the data
c..n      = number of data points

c..output:
c..x0        = x coordinate of the ellipse center
c..y0        = y coordinate of the ellipse center
c..xfoci1    = x coordinate of foci1 
c..yfoci1    = y coordinate of foci1 
c..xfoci2    = x coordinate of foci2 
c..yfoci2    = y coordinate of foci2 
c..rmaj      = length of semi-major axis
c..rmin      = length of semi-minor axis
c..foci_dist = distance from center to foci
c..tilt      = angle between the major axis and the x-axis
c..eccen     = orbital eccentricity
c..a         = elipse equation coefficient
c..b         = elipse equation coefficient
c..c         = elipse equation coefficient
c..d         = elipse equation coefficient
c..e         = elipse equation coefficient
c..f         = elipse equation coefficient
c..asig      = uncertainty in a, a +/- asig
c..bsig      = uncertainty in b, b +/- bsig
c..csig      = uncertainty in c, c +/- csig
c..dsig      = uncertainty in d, d +/- dsig
c..esig      = uncertainty in e, e +/- esig
c..fsig      = uncertainty in f, f +/- fsig
c..chisq     = chi-square per degree of freedom


c..ellipse:  (1+a) x**2 + (1-a) y**2 + c*x*y + d*x + e*y + f = 0
c..    x**2 + y**2 + a*(x**2 - y**2) + c*x*y + d*x + e*y + f = 0
c..                  a*x**2 + b*y**2 + c*x*y + d*x + e*y + f = 0
c..
c..the second form is used for the fitting, and the coefficients
c..for the third form are returned.


c..declare
       integer          n,i,nmax,npar
       parameter        (nmax=500, npar=5)
       double precision x(n),y(n),x0,y0,xfoci1,yfoci1,xfoci2,yfoci2, &
                       rmaj,rmin,foci_dist,tilt,eccen,a,asig,b,bsig, &
                       c,csig,d,dsig,e,esig,f,fsig,                  &
                       u(nmax,npar),w(npar),v(npar,npar),bp(nmax),   &
                       ap(npar),cvm(npar,npar),wmax,thresh,sum,chisq,&
                       det,aaa,bbb,fff,den,t1,t2,                    &
                       tol,tiny,small
       parameter        (tol = 1.0d-12, tiny = 1.0d-99, small=1.0d-14)


c..bullet check
      if (n .gt. nmax) stop 'n > nmax in routine fitell'


c..set the u matrix and right hand side
c..the entries of u are simply the terms in front of the parameters to be fit
c..and the right hand side is everything else moved across the equal sign
      do i = 1,n
       u(i,1) = x(i)**2 - y(i)**2
       u(i,2) = x(i)*y(i)
       u(i,3) = x(i)
       u(i,4) = y(i)
       u(i,5) = 1.0d0
       bp(i)  = -(x(i)**2 + y(i)**2)
      enddo


c..singular value decomposition
c..a call to a double precision version of the numerical recipes routine

      call svdcmp(u,n,npar,nmax,npar,w,v)


c..massage the singular values
      wmax = 0.0d0
      do i=1,3
       if (w(i) .gt. wmax) wmax = w(i)
      enddo
      thresh = tol * wmax
      do i=1,3
       if (w(i) .lt. thresh) w(i) = 0.0d0
      enddo

c..backsubstitution and covariance matrix
c..return the uncertainties in the fitted parameters sqrt(cvm(i,i)) if desired
c..calls to double precision versions of the numerical recipes routines

      call svbksb(u,w,v,n,npar,nmax,npar,bp,ap)
      call svdvar(v,npar,npar,w,cvm,npar)


c..evaluate the chi-square per degree of freedom
      chisq = 0.0d0
      do i=1,n
       sum = ap(1)*(x(i)**2 - y(i)**2) + ap(2)*x(i)*y(i) + ap(3)*x(i)  &
            + ap(4)*y(i) + ap(5)
       chisq = chisq + (bp(i) - sum)**2
      enddo
      chisq = chisq/(n - npar)


c..set the return parameters
c..coefficients of the ellipse equation
      a    = 1.0d0 + ap(1)
      b    = 1.0d0 - ap(1)
      c    = ap(2)
      d    = ap(3)
      e    = ap(4)
      f    = ap(5)


c..and their uncertainties
      asig = sqrt(cvm(1,1))
      bsig = sqrt(cvm(1,1))
      csig = sqrt(cvm(2,2))
      dsig = sqrt(cvm(3,3))
      esig = sqrt(cvm(4,4))
      fsig = sqrt(cvm(5,5))


c..center of the ellipse
      det  = 4.0d0 * a * b - c * c + tiny
      if (det .eq. 0) stop 'bad determinant in routine fitell'
      x0   = (c*e - 2.0d0*b*d) / det
      y0   = (c*d - 2.0d0*a*e) / det
      den  = a - b
      if (abs(den) .lt. small) den = 0.0d0


c..angle between the major axis and the x-axis
c..an angle between 0 and 180 degrees
      tilt = 90.0d0 - 0.5d0 * atan2d(-c, den)


c..semi-major and semi-minor axes
      aaa = 0.5d0 * (a + b + den/cosd(2.0d0*tilt))
      bbb = a + b - aaa
      fff = a *x0*x0 + b*y0*y0 + c*x0*y0 - f
      if (fff/aaa .lt. 0.0   .or.  fff/bbb .lt. 0.0) &
         stop 'not an ellipse in routine fitell'
      t1   = sqrt(fff/aaa)
      t2   = sqrt(fff/bbb)
      rmaj = max(t1,t2)
      rmin = min(t1,t2)


c..distance from center to the foci and the orbital eccentricity
      foci_dist = sqrt(rmaj**2 - rmin**2)
      eccen     = foci_dist/rmaj


c..coordinates of the two foci
      xfoci1 = x0 + cosd(tilt) * foci_dist 
      yfoci1 = y0 + sind(tilt) * foci_dist

      xfoci2 = x0 - cosd(tilt) * foci_dist 
      yfoci2 = y0 - sind(tilt) * foci_dist

      return
      end





      subroutine ranmar(ijkl,rvec,len) 
      implicit none 
      save

c..marsaglia and zaman random number generator. 
c..period is 2**43 with 900 million different sequences
c..the state of the generator (for restarts) is in the common block. 

c..input:
c..ijkl = seed set to negative integer to initialize or reinitialize 
c..rvec = array
c..len  = length of rvec

c..output:
c..ijkl = set to +ijkl after initialization
c..rvec = array of random numbers between 0.0 and 1.0

c..ijkl = 54217137 is the same as the four seeds i=12, j=34, k=56, l=78 of 
c..marsagalia and zaman 

c..declare 
      integer          ijkl,len,ivec,ij,kl,i,j,k,l,ii,jj,m,iff 
      double precision rvec(len),uni,s,t 

      integer          i97,j97 
      double precision u(97),c,cd,cm 
      common /ranset1/ u,c,cd,cm,i97,j97 

      data             iff/0/ 


c..initalization initialize even if the user forgot to set ijkl to
c..a negative value the first time.
      if (ijkl .lt. 0 .or. iff .eq. 0) then 
       iff  = 1 
       if (ijkl .lt. 0) ijkl = -ijkl 
       ij   = ijkl/30082 
       kl   = ijkl - 30082 * ij 
       i    = mod(ij/177,177) + 2 
       j    = mod(ij,177) + 2 
       k    = mod(kl/169,178) + 1 
       l    = mod(kl,169) 

       do ii=1,97 
        s = 0.0d0 
        t = 0.5d0 
        do jj =1,24 
         m = mod(mod(i*j,179)*k,179) 
         i = j 
         j = k 
         k = m 
         l = mod(53*l + 1, 169) 
         if (mod(l*m,64) .ge. 32) s = s + t 
         t = 0.5d0 * t 
        enddo
        u(ii) = s 
       enddo
       c   = 362436.0d0/16777216.0d0 
       cd  = 7654321.0d0/16777216.0d0 
       cm  = 16777213.0d0/16777216.0d0 
       i97 = 97 
       j97 = 33 
      end if 

c..normal execution starts here
      do ivec=1,len 
       uni = u(i97) - u(j97) 
       if (uni .lt. 0.0) uni = uni + 1.0d0 
       u(i97) = uni 
       i97 = i97 - 1 
       if (i97 .eq. 0) i97 = 97 
       j97 = j97 - 1 
       if (j97 .eq. 0) j97 = 97 
       c = c - cd 
       if (c .lt. 0.0) c = c + cm 
       uni = uni - c 
       if (uni .lt. 0.0) uni = uni + 1.0d0 
       rvec(ivec) = uni 
      enddo
      return 
      end 




      subroutine svdcmp(a,m,n,mp,np,w,v) 
      implicit none 
      save

c..given a matrix a, logical dimensions m by n, physical dimension mp by np, 
c..this routine computes its singular value decomposition 
c..a = u * w * vtranspose 

c..the matrix u replaces a on output. The diagonal of singular values 
c..is output as the vector w. the matrix v, not its transpose, is also output.
c..m must be greater than or equal to n; if it is smaller, then a should 
c..be filled up to a square with zero rows 


c..declare 
      integer          m,n,mp,np,i,j,k,l,its,nmax,nm,jj 
      parameter        (nmax = 500) 
      double precision a(mp,np),w(np),v(np,np),rv1(nmax),scale,  & 
                      anorm,c,f,g,h,s,x,y,z,pythag 

c..householder reduction to bidiagonal form 
      g     = 0.0d0 
      scale = 0.0d0 
      anorm = 0.0d0 
      do i=1,n    
       l      = i + 1    
       rv1(i) = scale*g   
       g      = 0.0d0 
       s      = 0.0d0    
       scale  = 0.0d0    
       if (i .le. m) then 
        do k=i,m    
         scale = scale + abs(a(k,i))  
        enddo
        if (scale .ne. 0.0) then 
         do k=i,m  
          a(k,i) = a(k,i)/scale    
          s      = s + a(k,i)*a(k,i)  
         enddo
         f      = a(i,i) 
         g      = -sign(sqrt(s),f)   
         h      = f*g - s  
         a(i,i) = f - g   

         do j=l,n    
          s = 0.0d0 
          do k=i,m  
           s = s + a(k,i)*a(k,j)  
          enddo
          f = s/h    
          do k=i,m  
           a(k,j) = a(k,j) + f*a(k,i) 
          enddo
         enddo

         do k=i,m 
          a(k,i) = scale*a(k,i)    
         enddo
        end if  
       end if    
       w(i)  = scale *g    
       g     = 0.0d0 
       s     = 0.0d0  
       scale = 0.0d0    
       if ((i .le. m) .and. (i .ne. n)) then  
        do k=l,n    
         scale = scale + abs(a(i,k))  
        enddo
        if (scale .ne. 0.0) then 
         do k=l,n  
          a(i,k) = a(i,k)/scale    
          s      = s + a(i,k)*a(i,k)  
         enddo

         f      = a(i,l) 
         g      = -sign(sqrt(s),f)   
         h      = f*g - s  
         a(i,l) = f - g   
         do k=l,n  
          rv1(k) = a(i,k)/h    
         enddo

         do j=l,m    
          s = 0.0d0 
          do k=l,n  
           s = s + a(j,k)*a(i,k)  
          enddo
          do k=l,n  
           a(j,k) = a(j,k) + s*rv1(k) 
          enddo
         enddo

         do k=l,n  
          a(i,k) = scale*a(i,k)    
         enddo
        end if  
       end if    
       anorm = max(anorm,(abs(w(i))+abs(rv1(i)))) 

c..end of householder reduction to bidiagonal form 
      enddo


c..accumulation of the right hand transformations 
      do i=n,1,-1 
       if (i .lt. n) then 

c..double division to avoid underflow 
        if (g .ne. 0.0) then 
         do j=l,n  
          v(j,i) = (a(i,j)/a(i,l))/g   
         enddo
         do j=l,n  
          s = 0.0d0 
          do k=l,n    
           s = s + a(i,k)*v(k,j)    
          enddo
          do k=l,n    
           v(k,j) = v(k,j) + s*v(k,i)   
          enddo
         enddo
        end if  

        do j=l,n    
         v(i,j) = 0.0d0 
         v(j,i) = 0.0d0 
        enddo
       end if    
       v(i,i) = 1.0d0 
       g      = rv1(i) 
       l      = i  

c..end accumulation of the right hand transformations 
      enddo


c..accumulate the left hand transforms 
      do i=min(m,n),1,-1 
       l = i+1    
       g = w(i)   
       do j=l,n    
        a(i,j) = 0.0d0 
       enddo

       if (g .ne. 0.0) then   
        g=1.0d0/g    
        do j=l,n  
         s = 0.0d0 
         do k=l,m    
          s = s + a(k,i)*a(k,j)    
         enddo
         f = (s/a(i,i))*g 
         do k=i,m    
          a(k,j) = a(k,j) + f*a(k,i)   
         enddo
        enddo
        do j=i,m    
         a(j,i) = a(j,i)*g  
        enddo

       else 
        do j= i,m   
         a(j,i) = 0.0d0 
        enddo
       end if    
       a(i,i) = a(i,i) + 1.0d0 

c..end accumulate the left hand transforms 
      enddo


c..diagonalization of the bidiagonal form 
c..loop over singular values 
      do k=n,1,-1 

c..loop over allowed iterations 
       do its=1,30   

c..test for splitting is here, note rv(1) is always zero 
        do l=k,1,-1 
         nm = l-1   
         if ((abs(rv1(l))+anorm) .eq. anorm) goto  2   
         if ((abs(w(nm))+anorm)  .eq. anorm) goto  1    
        enddo

c..cacellation of rv1(l) if l .gt. 1 
1       c = 0.0d0 
        s = 1.0d0 
        do i=l,k    
         f      = s * rv1(i)   
         rv1(i) = c * rv1(i) 
         if ((abs(f)+anorm) .eq. anorm) goto  2 
         g    = w(i) 
         h    = pythag(f,g)  
         w(i) = h 
         h    = 1.0d0/h    
         c    = (g*h)   
         s    = -(f*h)   
         do j=1,m    
          y       = a(j,nm)    
          z       = a(j,i) 
          a(j,nm) = (y*c) + (z*s)  
          a(j,i)  = -(y*s) + (z*c)  
         enddo
        enddo

c..convergence check 
2       z = w(k) 
        if (l .eq. k) then   
         if (z .lt. 0.0) then   
          w(k) = -z    
          do j=1,n    
           v(j,k) = -v(j,k)   
          enddo
         end if    
         goto  3  
        end if  
        if (its .eq. 30) stop 'no convergence in 30 iters in svdcmp' 

c..shift the bottom 2 by 2 minor 
        x  = w(l) 
        nm = k-1 
        y  = w(nm)    
        g  = rv1(nm)  
        h  = rv1(k)   
        f  = ((y-z)*(y+z)+(g-h)*(g+h))/(2.0d0*h*y)  
        g  = pythag(f,1.0d0) 
        f  = ((x-z)*(x+z)+h*((y/(f+sign(g,f)))-h))/x  


c..now the qr transformation 
        c = 1.0d0 
        s = 1.0d0 
        do j=l,nm   
         i      = j+1    
         g      = rv1(i) 
         y      = w(i)   
         h      = s*g    
         g      = c*g    
         z      = pythag(f,h) 
         rv1(j) = z 
         c      = f/z    
         s      = h/z    
         f      = (x*c) + (g*s)   
         g      = -(x*s) + (g*c)   
         h      = y*s    
         y      = y*c    
         do jj=1,n 
          x       = v(jj,j)  
          z       = v(jj,i)  
          v(jj,j) =  (x*c)+(z*s)   
          v(jj,i) = -(x*s)+(z*c)   
         enddo
         z    = pythag(f,h) 
         w(j) = z   

c..the rotation can be arbitrary if z is zero 
         if (z .ne. 0.0) then   
          z = 1.0d0/z    
          c = f*z  
          s = h*z  
         end if    
         f =  (c*g)+(s*y)   
         x = -(s*g)+(c*y)   
         do jj=1,m 
          y       = a(jj,j)  
          z       = a(jj,i)  
          a(jj,j) = (y*c)+(z*s)   
          a(jj,i) =-(y*s)+(z*c)   
         enddo

c..end of qr transform 
        enddo
        rv1(l) = 0.0d0 
        rv1(k) = f   
        w(k)   = x 

c..end of iteration loop
       enddo
3      continue 

c..end of loop over singular values 
      enddo
      return 
      end    




      double precision function pythag(a,b) 
      implicit none 
      save

c..computes sqrt(a**2 + b**2) without destructive underflow or overflow 

c..declare 
      double precision a,b,absa,absb 
      absa = abs(a) 
      absb = abs(b) 
      if (absa .gt. absb) then 
       pythag = absa * sqrt(1.0d0 + (absb/absa)**2) 
      else 
       if (absb .eq. 0) then 
        pythag = 0.0d0 
       else 
        pythag = absb * sqrt(1.0d0 + (absa/absb)**2) 
       end if 
      end if 
      return 
      end 





      subroutine svbksb(u,w,v,m,n,mp,np,b,x) 
      implicit none 
      save

c..solves ax=b for x when a is specified by the arrays u, w and v as returned 
c..by routine svdcmp. a is of logical dimension m by n and physical dimension 
c..mp by np. b is the input right hand side. x on output is the solution 
c..vector. no input quantities are destoyed, so one may call sequentially 
c..with different b's 

c..declare 
      integer          m,n,mp,np,nmax,i,j,jj 
      parameter        (nmax = 500) 
      double precision u(mp,np),w(np),v(np,np),b(mp),x(np),tmp(nmax),s 

c..compute utranspose * b 
      do j=1,n    
       s = 0.0d0 

c..nonzero result only if wj is nonzero; and divide 
       if (w(j) .ne. 0.0) then   
        do i=1,m    
         s = s + u(i,j)*b(i)  
        enddo
        s = s/w(j)   
       end if    
       tmp(j) = s 
      enddo


c..and matrix multiply to get the answer 
      do j=1,n    
       s = 0.0d0 
       do jj=1,n 
        s = s + v(j,jj)*tmp(jj)    
       enddo
       x(j) = s   
      enddo
      return 
      end    




      subroutine svdvar(v,ma,np,w,cvm,ncvm)  
      implicit none 
      save

c..evaluates the covariance matrix cvm of the fit for ma parameters obtained 
c..by the routine svdfit, call this routine with matrices v and w as returned 
c..by svdfit. np and ncvm are the physical dimensions of v,w and cvm. 

c..declare 
      integer          ma,np,ncvm,mmax,i,j,k 
      parameter        (mmax=20)    
      double precision v(np,np),w(np),cvm(ncvm,ncvm),wti(mmax),sum  

c..initialize 
      do i=1,ma   
       wti(i) = 0.0d0    
       if (w(i) .ne. 0.0) wti(i) = 1.0d0/(w(i)*w(i)) 
      enddo

c..sum the contributions to the covariant matrix; recipe equation 14.3.20 
      do i=1,ma   
       do j=1,i  
        sum = 0.0d0 
        do k=1,ma   
         sum = sum + v(i,k)*v(j,k)*wti(k) 
        enddo
        cvm(i,j) = sum   
        cvm(j,i) = sum   
       enddo
      enddo
      return 
      end    







