// ESA (C) 2000-2018 
//
// This file is part of ESA's XMM-Newton Scientific Analysis System (SAS).
//
//    SAS is free software: you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation, either version 3 of the License, or
//    (at your option) any later version.
//
//    SAS is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with SAS.  If not, see <http://www.gnu.org/licenses/>.
#ifndef STATISTICS_H
#define STATISTICS_H

#include <vector>
#include <iostream>
#include <cmath>
#include <algorithm>
#include <numeric>
#include <iomanip>
#include <cstring>
//#include "polint.h"
//#include "leastsquares.h"
#include "utility.h"
#include "errstr.h"
 
#include <f90cxx.h>




template<class kind>
vector<kind> resampleData(vector<kind> &data);
template<class kind1, class kind2>
vector<kind1> resampleData(vector<kind1> &data, kind2 n1);


using namespace std;


/***************************************************************
*
*
****************************************************************/
template<class kind>
kind Median(vector<kind> &data1, bool sorted = false, bool nulls = true)
{


    vector<kind> data;
    if (nulls)
    {
        for(unsigned i = 0; i < data1.size(); i++)
            if( !isNull(data1[i]) ) data.push_back( data1[i] );
    }
    else
        data = data1;


    if(!sorted)
    {
        sort(data.begin(), data.end());
    }



    double median;
    int nData = data.size();

    if (nData == 0) return (kind)0.0;
    if (nData == 1) return (kind)data[0];

    int n2 = nData / 2;

    if ( n2 * 2 == nData ) // Even number of points
        median = ((double)data[n2 - 1] + (double)data[n2]) / 2.0;
    else
      median = (kind)data[n2];

    return (kind)median;
}




/**************************************************************
*
*
***************************************************************/

template<class kind>
class MadVal
{

 public:
     vector<kind> _data;
     kind _median, _mad;
     vector<kind> _deviation;
     vector<kind> _badValue;
     kind _nSigma;
     MadVal( vector<kind> data ) : _data(data)
     {

     
         if( data.size() < 2)
         {
             _median = _data[0];
             _mad = 0.0;

         }
         else
         {
             vector<kind> temp = data;
             sort(temp.begin(), temp.end() );

             _median = Median(temp, true);

             int n = (int)data.size();
             

             for(int i = 0; i < n; i++)
             {
                 _deviation.push_back( abs(temp[i] - _median));


             }

             _mad = Median(_deviation);
             if( abs(_mad) < 0.0001)
	     {
	         _mad = 0.0001;

             } 
         }
     };
     vector<kind> test(kind nSigma = 5)
     {
         _nSigma = nSigma;
         _badValue.clear();
         vector<kind> newData;
         for(unsigned i = 0; i < _deviation.size(); i++)
	 {

           double df = abs(_data[i] - _median);
	   if ( df / _mad < nSigma)
	   {
	       newData.push_back( _data[i] );
             
	   } 
           else
	   {
	      _badValue.push_back( _data[i] );

           }
         }
         return newData;
 
	

     };

     void listBadValues(string caption)
     {

       if(_badValue.size() == 0) return;
    

       errstr << caption << " (nsigma=" << setprecision(1) <<  setiosflags(ios::fixed) 
              << _nSigma << setprecision(2) << ", median value=" << _median << ")" 
              << Sas::message(Msg::App, Msg::Sparse);


      


       for(unsigned i = 0; i < _badValue.size(); i++)
       {
      
           errstr << i << ") x=" << setprecision(2) << setiosflags(ios::fixed) << _badValue[i]
                  << Sas::message(Msg::App, Msg::Sparse);



       }  


     }

};

/********************************************************************
*
*
********************************************************************/
template<class kind>
class MEstimator
{

 public:
     vector<kind> _data;
     kind _median, _mad;
     vector<kind> _deviation;
     vector<kind> _badValue;
     kind _nSigma;

     MEstimator( vector<kind> data ) : _data(data)
     {

   

     }
     void computeT()
     {
       int maxIter = 10;
       vector<double> T(maxIter), S(maxIter);


       MadVal<kind> mad( _data );
  
       T[0] = mad._median ;
       S[0] = mad._mad;
       for(int iter = 1; iter < maxIter; iter++)
       {
	 // T[i] = T[i - 1] + computeM(;    


       }
    }
    /****************************************************
    * 
    *
    ****************************************************/
    kind computeM(kind T, kind S)
    {
        double sum = 0.0;
        
        for(unsigned i = 0; i < _data.size(); i++)
	{
	    sum += ( (double)_data[i] - (double)T) / (double)S;  

        }


       

     }


};



/********************************************************************
*
*
*********************************************************************/
template<class kind>
kind MinVal(vector<kind> &data)
{
    kind minData = data[0];
    for(unsigned i = 0; i < data.size(); i++)
      minData = min(minData, data[i]);

    return minData;



}

/********************************************************************
*
*********************************************************************/
template<class kind>
kind MaxVal(vector<kind> &data)
{
    kind maxData = data[0];
    for(unsigned i = 0; i < data.size(); i++)
      maxData = max(maxData, data[i]);

    return maxData;

}






/********************************************************************
*
*********************************************************************/
template<class kind>
kind MinVal(kind *data, int size)
{
    kind minData = data[0];
    for(int i = 0; i < size; i++)
      minData = min(minData, data[i]);

    return minData;

}



/********************************************************************
*
*********************************************************************/
template<class kind>
kind MaxVal(kind *data, int size)
{
    kind maxData = data[0];
    for(int i = 0; i < size; i++)
      maxData = max(maxData, data[i]);

    return maxData;

}






/***************************************************************
*
*
****************************************************************/
template<class kind>
kind Mean(kind *data, long nData)
{

    double mean = 0.0;
    if (nData == 0) return 0.0;

    for( long i = 0; i < nData; i++)
        mean += (double)data[i];

    return (kind)(mean / (double)nData);

}




/****************************************************************
*
*
****************************************************************/
template<class kind>
kind Median(vector<vector< kind> > &data1)
{

    vector<kind> data;
    for(unsigned i = 0; i < data1.size(); i++)
        for(unsigned j = 0; j < data1[0].size(); j++)
	  if (!isNull( data1[i][j] ) )
              data.push_back(data1[i][j]);

    return Median(data, false, false);

  

}

/***************************************************************
*
*
****************************************************************/
template<class kind>
kind Mean(vector<kind> &data1, bool nulls = true)
{







    if (data1.size() == 0) return (kind)0.0;





    

    vector<kind> data;
    if (nulls)
    {
        for(unsigned i = 0; i < data1.size(); i++)
            if( !isNull(data1[i]) ) data.push_back( data1[i] );
    }
    else
        data = data1;

   




    sort(data.begin(), data.end());




    return (kind)(accumulate( data.begin(), data.end(), 0.0 ) / double(data.size() ));


}





/***************************************************************
*
*
****************************************************************/
template<class kind>
kind Median(kind *data, long nData)
{
    kind median;


    if (nData == 0) return 0.0;
    if (nData == 1) return data[0]; 



  
    int n2 = nData / 2;

    if ( n2 * 2 == nData ) // Even number of points
        median = (data[n2 - 1] + data[n2]) / 2.0;
    else
      median = data[n2];

    return median;
}    







/***************************************************************
*
*
****************************************************************/
template<class kind1, class kind2>
kind2 Percentile(vector<kind1> &data1, kind2 perCent, bool sorted = false, bool nulls = true)
{



   

    vector<kind1> data;
    if (nulls)
    {
        for(unsigned i = 0; i < data1.size(); i++)
            if( !isNull(data1[i]) ) data.push_back( data1[i] );
    }
    else
        data = data1;

   



    if(!sorted)
    {
        sort(data.begin(), data.end());
    }





    long size = data.size();
    
    long index = (long)(perCent * (float)size + 0.00001) - 1;
    index = max((long)0, index);
    index = min(index, size - 1); 
     
    //    cout << "n=" << size << " percent=" << perCent << " index=" << index << " percentile=" << data[index] << endl;
    //    cout << "data[0]=" << data[0] << " data[" << (size - 1) << "]" << data[size - 1] << endl;
    return (kind2)data[index];

}





/***************************************************************
*
*
****************************************************************/
template<class kind1, class kind2>
kind2 Percentile(kind1 *data, long size, kind2 perCent)
{

 
    
    long index = (long)(perCent * (float)size + 0.00001) - 1;
    index = max((long)0, index);
    index = min(index, size - 1); 

    return (kind2)data[index];

}





/**********************************************
 *                                            *
 **********************************************/
template<class kind>
class Statistics
{
    private:
        vector<double> _data;
    public:
        kind _min;
        kind _max;
        kind _mean;
        kind _median;
        kind _mode;
        kind _p05;
        kind _p95;
        int _noNonNulls;
        int _size;
    public:
        
        Statistics(vector<kind> &data)
	{
	    initialise( data );

        }
	void initialise(vector<kind> & data)
        {

	    for(unsigned i = 0; i < data.size(); i++)
	        if ( !isNull(data[i]) ) _data.push_back( (double)data[i] );

            

            /*****************************************************
	    * Sort the data
	    *****************************************************/

            _size = _data.size();

            if(_size == 0) 
	    {
	        _min = _max = _median = _mean = _mode = 0.0;
                _p05 = _p95 = 0.0;
                _noNonNulls = 0;
                return;


            }
	    std::sort( _data.begin(), _data.end() );
            _min = _data[0];

            _max = _data[ _size - 1 ];

            _mean = Mean(_data, false );
            _median = Median(_data, true, false);

	    //            double mean, std, median;


	    //	    _mode = (kind)halfRangeMode( _data, mean, std, median);

	    _mode = 0.0;



            _p05 = Percentile(_data, 0.05, true, false);
            _p95 = Percentile(_data, 0.95, true, false);

            _noNonNulls = _data.size(); 

         

        }
        Statistics() {};

};





















F90MODPROCPROTO(double, omfunctions, gammaq)(double*, double*);

template<class kind1, class kind2>
kind2 halfRangeMode(vector<kind1> & data, kind2 &mean, kind2 &std, kind2 &median);

template<class kind1, class kind2>
void maxInterval(vector<kind1> &data, long unsigned &iStart, long unsigned &iFinish, kind2 constant = 0.5);




template<class kind1, class kind2>
kind2 modeError(vector<kind1> &data, kind2 &mode);



/********************************************************
*
********************************************************/
template<class kind1, class kind2>
kind2 halfRangeMode(vector<kind1>& data, kind2 &mean, kind2 &std, kind2 &median)
{

  
    sort( data.begin(), data.end() );


    vector<kind1> tempData1 = data;

    std = 0.0;


    median = (kind2)Median( tempData1, true, false );

    median = max( (kind2)2.0, median );




    mean   = (kind2)Mean( data );

 

    
    unsigned long minSize = max((long)5, (long)data.size() / 20);

       
   
    
    unsigned size = tempData1.size();
    if (size < minSize) return Mean(data);
    


    median = Median(tempData1);
    mean   = Mean(tempData1);

    long unsigned iStart = 0, iFinish = tempData1.size() - 1;  
    
   

    for (int i = 1; i <= 10000; i++)
    {

      

      //      cout << "Half-rnage mode - iteration " << i << endl;   


        maxInterval(tempData1, iStart, iFinish, 0.5);


   
        unsigned long size1 = iFinish - iStart + 1;
      

	if (size <= minSize || size == size1)    
        {


	    vector<float> tempData;
            for(unsigned j = iStart; j <= iFinish; j++)
	        tempData.push_back( tempData1[j]);

	    float mean = Mean( tempData ); 	 
      
	    //           cout << "Half-range mode - istart=" << iStart << " ifinish=" << iFinish << " mean=" << mean << endl;

	  
            

	    //            float std = 3.0 * max(0.5, (double)StandardDeviation( tempData ));
          
	    //            float xl = mean - std;
	    //            float xu = mean + std;

	    //            cout << "The final data\n";
	    //	    for (unsigned long i = 0; i < tempData.size(); i++)
	    //	      cout << i << ") data=" << tempData[i] << endl;	    

          
	    //            iStart += tempData.size();

	    //            for (unsigned long i = iStart; i < data.size(); i++)
	    //	        if ( data[i] <= xu) tempData.push_back(data[i]);

	    //	    mean = Mean( tempData );

	    //	    cout << "half-range mode- initila mean=" << mean << " size=" << size << " istart=" << iStart << " final mean=" << finalMean << endl;
	   
	    //            cout << "Returning mean " << mean << endl; 
            return mean;    

        }
        
        size = size1;

    
   
	
 
    };
    cout << "Failed to finish!\n";
    cout << "size=" << size << endl;
    exit(0);
}
/****************************************************
*
*
****************************************************/
template<class kind1, class kind2>
void maxInterval(vector<kind1> &data, long unsigned &iStart, long unsigned &iFinish, kind2 constant)
{

  
    vector< vector<kind1> > bin;
    long unsigned n = iStart - iFinish + 1; 

 
    

    long unsigned iStart1 = iStart, iFinish1 = iFinish;
    kind1 width = data[iFinish] - data[iStart];

    if (width < 0.1) return;

    kind1 halfWidth = width * (kind1)constant;


    kind1 wMin = (kind1)10000000000.0;    

    long unsigned maxCount = 0;
    long unsigned noBins = 0; 

 
  
    long unsigned iu = iStart;

    for(unsigned long i = iStart; i <= iFinish; i++)
    {
    
   
        kind1 f = data[i];
       
	kind1 fPlus  = f + halfWidth;
	//		cout << "Maxinterval i=" << i << " f=" << f << " fplus=" << fPlus << endl;    
 
     
        unsigned iu1 = iFinish;
        for(unsigned j = iu; j < iFinish; j++)
        {
	    if (data[j] > fPlus)
	    {
	        iu1 = j - 1;
                break;
            }   
      
        } 
	iu = iu1;
	

	//	cout << "Line 661 istart1=" << iStart1 << " iu=" << iu << endl;
        long unsigned count = iu - i + 1;
      	 
	if (count == n)
        {
          iStart = iStart1;
          iFinish = iFinish1;
	  return;
        
        }
        if (count > maxCount)
	{ 
	 
	    iStart1 = i;
            iFinish1 = iu;
            wMin = data[iu] - data[i];
   
	    //         	cout << "Line 678 istart1=" << iStart1 << " iu=" << iu << endl;
   
       
            noBins = 1;
	  
            maxCount = count;   
        }
        else if (count == maxCount)
	{
	


	    kind1 width = data[iu] - data[i];
            if (width < wMin)
	    {
	        wMin = width;   // Smallest bin width  

                iStart1 = i;
                iFinish1 = iu;
     
		//              	cout << "Line 698 istart1=" << iStart1 << " iu=" << iu << endl;
   
                noBins = 1;

            }  
            else if( width > wMin)
	    {
	      continue;
            }
            else
	    {


	        iFinish1 = iu;
       
		//	cout << "Line 713 istart1=" << iStart1 << " iu=" << iu << endl;
   

                noBins++;
	    }
      

          
        }
    } 

    //   cout << "Line 710 noBins=" << noBins << " istart1=" << iStart1 << " ifinish1=" << iFinish1 << endl;


    if (noBins == 1)
    {     

        iStart = iStart1;
        iFinish = iFinish1;
        return;
    }


       
   

    if ( (data[iStart1 + 1] - data[iStart1]) < (data[iFinish1] - data[iFinish1 - 1]) )
    {
        
	iFinish1--;
    }
    else if ( (data[iStart1 + 1] - data[iStart1]) > (data[iFinish1] - data[iFinish1 - 1]))
    {
       
      
        iStart1++;
    }
    else if ( (data[iStart1 + 1] - data[iStart1]) == (data[iFinish1] - data[iFinish1 - 1]))
    {
       
      
      
        iStart1++;
        iFinish1--;
    }


   
    //        cout << "Line 788\n";
			   
         
    iStart = iStart1;
    iFinish = iFinish1;


}




/*****************************************************************
*
*****************************************************************/
template<class kind1, class kind2>
kind2 modeError(vector<kind1> &data, kind2 &mode)
{

    unsigned long n = data.size() ;
    kind1 logN = log10((float)n);
    long unsigned nSims = n * unsigned(logN * logN);
 
    nSims = max(n, nSims);
    nSims = min(nSims, (long unsigned)100);
 
    vector<kind2> simMode;


   
    for (long unsigned i = 0; i < nSims; i++)
    {
 

   
        vector<kind1> tempData = resampledData(data);
      
	//        for(unsigned j = 0; j < data.size(); j++)
	//	  cout << j << ") resampled data=" << tempData[j] << endl;

     
        kind2 mean, std, median;
        kind2 mode = halfRangeMode(tempData, mean, std, median);
        simMode.push_back(mode);
	//	cout << "Simulation " << i << " mode=" << simMode[i] << endl;
      

    }

    mode = Mean(simMode);  

    sort(simMode.begin(), simMode.end());

    /************************************************************  
    * 95% confidence limit
    ************************************************************/
    long unsigned il = (int)floor(0.05 * float(nSims)) - 1;
    il = max(il, (unsigned long)1);
    long unsigned iu = (unsigned)ceil(0.95 * float(nSims)) - 1;
    iu = min(iu, nSims - (long unsigned)2);
    cout << "nsims=" << nSims << " il=" << il << " iu=" << iu << endl;
    cout << "95% Confidence limit = " << simMode[il] << "-" << simMode[iu] << endl;

    double sum = 0.0;
    for (long unsigned i = 0; i < nSims; i++)
    {
      //      cout << "mode=" << mode << " sim=" << simMode[i] << endl;

        double diff = double(simMode[i] - mode); 
        sum += (diff * diff);

       
    }
  
    kind2 error = 3.0 * sqrt(sum / double(nSims));

    sum = 0.0;
    long unsigned k = 0;
    for (long unsigned i = 0; i < nSims; i++)
    {
      
        if (fabs(simMode[i] - mode) <= error)
	{
            double diff = double(simMode[i] - mode); 
            sum += (diff * diff);
            k++;
        }
       
    }
   
    error = sqrt(sum / double(k));

    return error;



}










/*****************************************************************
*
*****************************************************************/
template<class kind1, class kind2>
void MonteErrors(vector<kind1> &data, kind2 &mean, kind2 &meanErr, 
                kind2 &meanP1, kind2 &meanP2, 
                kind2 &median, kind2 &medianErr, 
                kind2 &medianP1, kind2 &medianP2)
{



    unsigned  n = data.size() ;
    float logN = log10((float)n);
    unsigned nSims = n * unsigned(logN * logN);

    nSims = min((unsigned)10000, nSims); 
    sort(data.begin(), data.end());

    mean = (kind2)Mean(data);
    median = (kind2)Median(data);
   
 
    vector<kind1> mean1, median1;

  
   
    for (unsigned i = 0; i < nSims; i++)
    {
   
    
        vector<kind1> tempData = resampleData(data);
      
        // Compute the mean from this data set        
        mean1.push_back( Mean( tempData ));

	tempData.clear();
       
  
	tempData = resampleData( data, n / 2);


	
    

  
        // Compute the median from this data set        


        sort(tempData.begin(), tempData.end());
        median1.push_back( Median( tempData ));

   
    }
    
    double sum1 = 0.0, sum2 = 0.0;
    for(unsigned i = 0; i < nSims; i++)
    {
        double diff = mean - mean1[i];
        sum1 += diff * diff;
 
        diff = median - median1[i];
        sum2 += diff * diff;
 



    }  

    meanErr = (kind2)sqrt(sum1 / (double)nSims);




    sort(mean1.begin(), mean1.end());

    //    cout << "Input mean=" << mean << " median=" << median << endl;

    //    for(unsigned i = 0; i < mean1.size(); i++)
    //      cout << i << ") mean=" << mean1[i] << " median=" << median1[i] << endl; 

    float p = 0.95;

    meanP1 = (kind2)Percentile(mean1, 1.0 - p, true, false);
    meanP2 = (kind2)Percentile(mean1, p, true, false);


    sort(median1.begin(), median1.end());

    medianErr = (kind2)sqrt(sum2 / (double)nSims);
    medianP1 = (kind2)Percentile(median1, 1.0 - p, true, false);
    medianP2 = (kind2)Percentile(median1, p, true, false);

    //    exit(0);

}
/***********************************************************************
*
***********************************************************************/
template<class kind>
vector<kind> resampleData(vector<kind> &data)
{

  
   
    long unsigned n = data.size();
 
 
  

    vector<kind> tempData;
 
      
    for(unsigned i = 0; i < n; i++)
    {

        unsigned k = rand() % (n - 1);

     
      


	tempData.push_back( data[k] ); 

	//        cout << i << ") k=" << k << " data=" << data[k] << endl;



    }

    
    return tempData;
}

/***********************************************************************
*
***********************************************************************/
template<class kind1, class kind2>
vector<kind1> resampleData(vector<kind1> &data, kind2 n1)
{

    
    long unsigned n = data.size();
 
  
    vector<kind1> tempData;
   
    vector< bool > used(n, false);
      
    for(unsigned i = 0; i < (unsigned)n1; i++)
    {


        while(1) {
            unsigned k = rand() % (n - 1);

	    
            if (!used[k])
            { 
	  
                tempData.push_back( data[k] );
                used[k] = true;
	        break;
        
            }
        }    

     

    }
    
    return tempData;

}






#endif




































