// ESA (C) 2000-2018 
//
// This file is part of ESA's XMM-Newton Scientific Analysis System (SAS).
//
//    SAS is free software: you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation, either version 3 of the License, or
//    (at your option) any later version.
//
//    SAS is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with SAS.  If not, see <http://www.gnu.org/licenses/>.
#include "createimage.h"
#include <vector>
#include "Dal.h"
#include <iostream>
#include <cmath>
/*******************************************************
*
*
*******************************************************/
void addInt8Image(DataSet *set, const char *name, const vector< vector<int8> >& image)

{


    unsigned nx = image.size();
    unsigned ny = image[0].size();
    longVec size(2);
    size[0] = ny;
    size[1] = nx;
   
   
    cout << "Adding integer 8 quality image\n";
  
  
    Array * array  = set->addArray(name,
				   Dal::Int8,
				   size);
    ArrayData * arraydata = array->data();
    int8 * data = arraydata->int8Data();
 

    array->nullValue((long)127);

    cout << "nullvalue=" << array->intNullValue() << endl;

    unsigned dx = size[1];
    unsigned dy = size[0];

    for (unsigned x = 0; x < dx; ++x)
    {
      for (unsigned y = 0; y < dy; ++y)
      {
             
	  int index = x + y * dx; 
	  if( image[x][y] < 127)
	  {    
	        data[index] = image[x][y];
	  }
	  else
	  {
	      arraydata->setToNull(index);
          }
      }

    }

    delete arraydata;


}



/*******************************************************
*
*
*******************************************************/
void addInt16Image(DataSet *set, const char *name, const vector< vector<int16> >& image)

{


    unsigned nx = image.size();
    unsigned ny = image[0].size();
    longVec size(2);
    size[0] = ny;
    size[1] = nx;
   
   
    cout << "Adding integer 16 quality image\n";
  
  
    Array * array  = set->addArray(name,
				   Dal::Int16,
				   size);
    ArrayData * arraydata = array->data();
    int16 * data = arraydata->int16Data();
 

    array->nullValue((long)32767);

    cout << "nullvalue=" << array->intNullValue() << endl;

    unsigned dx = size[1];
    unsigned dy = size[0];

    for (unsigned x = 0; x < dx; ++x)
    {
      for (unsigned y = 0; y < dy; ++y)
      {
             
	  int index = x + y * dx; 
	  if( image[x][y] >= 0)
	  {    
	        data[index] = image[x][y];
	  }
	  else
	  {
	      arraydata->setToNull(index);
          }
      }

    }

    delete arraydata;


}

/*******************************************************
*
*
*******************************************************/
void addReal32Image(DataSet *set, const char *name, const vector< vector<real32> >& image)

{


    unsigned nx = image.size();
    unsigned ny = image[0].size();
    longVec size(2);
    size[0] = ny;
    size[1] = nx;
   
   
 
  
  
    Array * array  = set->addArray(name,
				   Dal::Real32,
				   size);
    ArrayData * arraydata = array->data();
    real32 * data = arraydata->real32Data();
 


    unsigned dx = size[1];
    unsigned dy = size[0];

    for (unsigned x = 0; x < dx; ++x)
    {
      for (unsigned y = 0; y < dy; ++y)
      {
             
	  int index = x + y * dx; 
	  if( image[x][y] >= 0)
	  {    
	        data[index] = image[x][y];
	  }
	  else
	  {
	      arraydata->setToNull(index);
          }
      }

    }

    delete arraydata;


}





/*******************************************************
*
*
*******************************************************/


vector<vector<int16> >  loadInt16Image(const string inSet, const char * name, int &nx, int &ny)
{


    DataSet *set = dataSetServer->open(inSet, Dal::Read);
 


    /*****************************************************************
    * Get the size of the image
    *****************************************************************/
    const Array* array = set->array(name);
 

    const ArrayData * arraydata = array->data();
    const int16 * theData = arraydata->int16Data();
 

    /*****************************************************************
    * Get the size of the image
    *****************************************************************/
 
    longVec size = array->size();
   



    nx = (int)size[1];
    ny = (int)size[0];


    vector< vector<int16> > image(nx); 


    for (int i = 0; i < nx; i++)
    {
        image[i].resize(ny); 
        for (int j = 0; j < ny; j++)
        {
            

	  image[i][j] = theData[i + (j * nx)];  
	 
	
        }

  
    }


    delete arraydata;
   
    dataSetServer->close( set ); // Finished with the file

    return image;
}



/*******************************************************
*
*
*******************************************************/


vector<vector<int16> >  loadInt16Image(const DataSet *set, const char * name, int &nx, int &ny)
{


 
 


    /*****************************************************************
    * Get the size of the image
    *****************************************************************/
    const Array* array = set->array(name);
 

    const ArrayData * arraydata = array->data();
    const int16 * theData = arraydata->int16Data();
 

    /*****************************************************************
    * Get the size of the image
    *****************************************************************/
 
    longVec size = array->size();
   



    nx = (int)size[1];
    ny = (int)size[0];


    vector< vector<int16> > image(nx); 


    for (int i = 0; i < nx; i++)
    {
        image[i].resize(ny); 
        for (int j = 0; j < ny; j++)
        {
            

	  image[i][j] = theData[i + (j * nx)];  
	 
	
        }

  
    }


    delete arraydata;
   
   

    return image;
}





/*******************************************************
*
*
*******************************************************/


vector<vector<float> >  loadFloatImage(const string inSet, const char * name, int &nx, int &ny)
{


    DataSet *set = dataSetServer->open(inSet, Dal::Read);
 


    /*****************************************************************
    * Get the size of the image
    *****************************************************************/
    const Array* array = set->array(name);
 

    const ArrayData * arraydata = array->data();
    const float * theData = arraydata->real32Data();
 

    /*****************************************************************
    * Get the size of the image
    *****************************************************************/
 
    longVec size = array->size();
   



    nx = (int)size[1];
    ny = (int)size[0];


    vector< vector<float> > image(nx); 


    for (int i = 0; i < nx; i++)
    {
        image[i].resize(ny); 
        for (int j = 0; j < ny; j++)
        {
            

	  image[i][j] = theData[i + (j * nx)];  
	 
	
        }

  
    }


    delete arraydata;
   
    dataSetServer->close( set ); // Finished with the file

    return image;
}



/*******************************************************
*
*
*******************************************************/


vector<vector<float> >  loadFloatImage(const DataSet *set, const char * name, int &nx, int &ny)
{


   
 


    /*****************************************************************
    * Get the size of the image
    *****************************************************************/
    const Array* array = set->array(name);
 

    const ArrayData * arraydata = array->data();
    const float * theData = arraydata->real32Data();
 

    /*****************************************************************
    * Get the size of the image
    *****************************************************************/
 
    longVec size = array->size();
   



    nx = (int)size[1];
    ny = (int)size[0];


    vector< vector<float> > image(nx); 


    for (int i = 0; i < nx; i++)
    {
        image[i].resize(ny); 
        for (int j = 0; j < ny; j++)
        {
            

	  image[i][j] = theData[i + (j * nx)];  
	 
	
        }

  
    }


    delete arraydata;
   
 

    return image;
}






/************************************************************
*
*
*
************************************************************/
void createImageFile(string file, vector< vector<float> > &image, int x0, int y0)
{

 
    DataSet *dataSet = dataSetServer->open(file, Dal::Create);


    int nx = (int)image.size();

    int ny = (int)image[0].size();


   *dataSet->addAttribute( "DATAMODE", "" ) = "IMAGING";
   *dataSet->addAttribute( "OBJECT", "" ) = "Chris";


   *dataSet->addAttribute( "FILTER", "" ) = "V";

   *dataSet->addAttribute( "OBS_ID", "" ) = "666";
   *dataSet->addAttribute( "EXP_ID", "" ) = "000000000000000";

   *dataSet->addAttribute( "DATE-OBS", "" ) = "2000-05-18T06:53:35.000";
   *dataSet->addAttribute( "DATE-END", "" ) = "2000-05-18T07:10:15.000";

   *dataSet->addAttribute( "EXPOSURE", "" ) = 1000.0;

   *dataSet->addAttribute( "WINDOWX0", "" ) = x0;
   *dataSet->addAttribute( "WINDOWDX", "" ) = ny * 2;


   *dataSet->addAttribute( "WINDOWY0", "" ) = y0;
   *dataSet->addAttribute( "WINDOWDY", "" ) = ny * 2;


   *dataSet->addAttribute( "BINAX1", "" ) = 1;
   *dataSet->addAttribute( "BINAX2", "" ) = 1;

    *dataSet->addAttribute( "BINBPE", "" ) = false;

   *dataSet->addAttribute( "DEADFRAC", "" ) = 0.0;
   *dataSet->addAttribute( "FRAMTIME", "" ) = 0.0;


    addReal32Image( dataSet, "PRIMARY", image);



    vector< vector<int16> > quality(nx, vector<int16>(ny, 0));

    addInt16Image( dataSet, "QUALITY", quality);



    dataSetServer->close( dataSet );



}

/****************************************************************
*
*
*****************************************************************/
int nint(float x)
{
    int i = int(x);
    if( x - float(i) > 0.5) ++i;
    return i;
}

/************************************************************
*
*
*
*************************************************************/
void addSource(vector<vector<float> > &image, float xPos, float yPos, float height)
{

    int nx = (int)image.size();
    int ny = (int)image[0].size();

    int dx = 6;
    int i0 = nint(xPos);
    int j0 = nint(yPos);

    int il = max(0, i0 - dx);
    int iu = min(nx - 1, i0 + dx);
    int jl = max(0, j0 - dx);
    int ju = min(ny - 1, j0 + dx);

    for(int i = il; i <= iu; i++)
    {
        float dx1 = float(i - i0);
        for(int j = jl; j <= ju; j++)
        {

            float dy1 = float(j - j0);
            float radius = hypot(dx1, dy1);
            if(radius <= float(dx) )
	    {
	        float f = height * exp(-radius * radius / 4.0);
                image[i][j] += f;
            }
        }
    }


}
